// Exceptions.hxx - exception definitions of murx
//
// (c) 2004 by Thomas Stauss, Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifndef MURX_EXCEPTIONS_HXX
#define MURX_EXCEPTIONS_HXX

#include <string>
#include "i18n.hxx"


namespace Murx
{ 
    namespace Exceptions
    {
        class Exception
        {
        protected:
            std::string exception_text;
        public:
            Exception () {}
            Exception (const std::string exception_text)
                : exception_text (exception_text)
                {
                }
            std::string what ()
                {
                    return exception_text;
                }
        };
        
        
        // >>> IOError
        class IOError : public Exception
        {
        public:
            IOError (const std::string exception_text)
                : Exception (exception_text)
                {
                }
        };
        
        
        class CouldNotOpenFile : public IOError
        {
        public:
            CouldNotOpenFile (const std::string exception_text)
                : IOError (_("could not open file ") + exception_text)
                {
                }
        };
        // <<< IOError
        
        
        class Network : public Exception
        {
        public:
            Network (const std::string exception_text)
                : Exception (exception_text)
                {
                }
        };
        
        // >>> Protocol
        class Protocol : public Exception
        {
        public:
            Protocol (const std::string exception_text)
                : Exception (exception_text)
                {
                }
        };
        
        class UnknownAccount : public Protocol
        {
        public:
            UnknownAccount ()
                : Protocol (_("account unknown by server"))
                {
                }
        };
        
        class AccountBlocked : public Protocol
        {
        public:
            AccountBlocked ()
                : Protocol (_("account is in use"))
                {
                }
        };
        
        class LoginError : public Protocol
        {
        public:
            LoginError (std::string message)
                : Protocol (_("wrong password or account in use"))
                {
                }
        };

        class CommandNotAvailable : public Protocol
        {
        public:
            CommandNotAvailable (std::string message)
                : Protocol (message)
                {
                }
        };

        class FolderDoesNotExist : public Protocol
        {
        public:
            FolderDoesNotExist (std::string folder)
                : Protocol (_("folder ") + folder + _(" does not exist") )
                {
                }
        };
        // <<< Protocol
        
        // >>> Network
        // >>> SSL
        class SSL : public Network
        {
        public:
            SSL (const std::string exception_text)
                : Network (exception_text)
                {
                }
        };
        
        
        class FingerprintChanged : public SSL
        {
        public:
            const std::string fingerprint_md5;
            const std::string fingerprint_sha1;
            
            FingerprintChanged (std::string fpr_md5, std::string fpr_sha1) 
                : SSL (_("fingerprint changed")),
                  fingerprint_md5  (fpr_md5),
                  fingerprint_sha1 (fpr_sha1)
                {
                }
        };
        
        
        class NoFingerprint : public SSL
        {
        public:
            const std::string fingerprint_md5;
            const std::string fingerprint_sha1;
            
            NoFingerprint (std::string fpr_md5, std::string fpr_sha1) 
                : SSL (_("no fingerprint available")),
                  fingerprint_md5  (fpr_md5),
                  fingerprint_sha1 (fpr_sha1)
                {
                }
        };
        
        class NoSSLContext : public SSL
        {
        public:
            NoSSLContext ()
                : SSL (_("could not initialize the SSL context"))
                {
                }
        };
        
        
        class SSLHandshake : public SSL
        {
        public:
            SSLHandshake ()
                : SSL (_("SSL handshake with server failed"))
                {
                }
        };
        // <<< SSL
        
        
        class NoSocket : public Network
        {
        public:
            NoSocket ()
                : Network (_("could not create socket"))
                {
                }
        };
        
        
        class NoConnection : public Network
        {
        public:
            NoConnection ()
                : Network (_("could not connect to server"))
                {
                }
        };
        
        class ConnectionLost : public Network
        {
        public:
            ConnectionLost ()
                : Network (_("lost connection to server"))
                {
                }
        };
        
        class Timeout : public Network
        {
        public:
            Timeout ()
                : Network (_("connection timed out"))
                {
                }
        };
        
        class NameResolution : public Network
        {
        public:
            NameResolution ()
                : Network (_("DNS entry not found"))
                {
                }
        };
        
        
        class UnknownHostname : public Network
        {
        public:
            UnknownHostname ()
                : Network (_("could not resolve hostname"))
                {
                }
        };
        // <<< Network
        
        
        
        // >>> Parser
        class Parser : public Exception
        {
        public:
            Parser (const std::string exception_text)
                : Exception (exception_text)
                {
                }
        };
        
        
        class AccountDefinitionIncomplete : public Parser
        {
        public:
            AccountDefinitionIncomplete (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class AccountNotFound : public Parser
        {
        public:
            AccountNotFound (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class DelimiterMissing : public Parser
        {
        public:
            DelimiterMissing (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class EmptyPattern : public Parser
        {
        public:
            EmptyPattern (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class EmptyString : public Parser
        {
        public:
            EmptyString (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class EnvironmentVariableNotSet : public Parser
        {
        public:
            EnvironmentVariableNotSet (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class InvalidFingerprint : public Parser
        {
        public:
            InvalidFingerprint (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class InvalidVariableIdentifier : public Parser
        {
        public:
            InvalidVariableIdentifier (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class LexicographicalError : public Parser
        {
        public:
            LexicographicalError (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class NameAlreadyExists : public Parser
        {
        public:
            NameAlreadyExists (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class RegExCompilationError : public Parser
        {
        public:
            RegExCompilationError (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class SyntaxError : public Parser
        {
        public:
            SyntaxError (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        
        
        class UnknownProtocol : public Parser
        {
        public:
            UnknownProtocol (const std::string exception_text)
                : Parser (exception_text)
                {
                }
        };
        // <<< Parser
        
    } /* namespace Exceptions */

} /* namespace Murx */

#endif /* MURX_EXCEPTIONS_HXX */
