// Filter.cxx - part of the murx project
//
// (c) 2004 by Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <string>
#include <vector>
#include <iostream>
#include "i18n.hxx"
#include "Exceptions.hxx"
#include "Filter.hxx"
extern "C"
{
#include <sys/types.h>
#if USE_PCRE
#  include <pcreposix.h>
#else
#  include <regex.h>
#endif
}
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


Murx::Filter::Filter ()
{
    // set default values
    minsize = 0;
    maxsize = 0;

    // default is not testing a filter
    testmode = false;
}


Murx::Filter::Filter (const char* expression, bool ignore_case,
                      bool negativity, bool only_check_body_lines)
{
    add_rule (expression,ignore_case,negativity,only_check_body_lines);

    // add new filter rule
    //this->expression.push_back  (expression);
    //this->ignore_case.push_back (ignore_case);
    //this->negativity.push_back  (negativity);
    
    // minimum and maximum size of a mail
    // 0 = off (default)
    minsize = 0;
    maxsize = 0;

    // default is not testing a filter
    testmode = false;

    // regex ist not compiled yet
    //compile_expression (expression.size ()-1);
}


Murx::Filter::~Filter ()
{
    // clear containers
/*
    for (unsigned int i = 0; i < rules.size (); i++)
    {
        delete rules[i];
    }
*/
}


// protected
void Murx::Filter::compile_expression (std::string expression, bool ignore_case,
                                       regex_t* comp_expression)
{
    int             comp_error;
    //regex_t            comp_expression;
    unsigned int    cflags = 0;

    // paramters for regcomp:
    //
    // regex_t* compiled_expression (container)
    // char*    expression
    // cflags   dcba (bits)
    //          a = REG_EXTENDED (1=yes)
    //          b = IGNORE_CASE  (1=yes)
    //          c = REG_NEWLINE  (1=no match of newline characters)
    //          d = REG_NOSUB    (1=only report success or failure and
    //                              no substrings)
    //
    //          ^--- the manpage of regcomp(3) is buggy because bits c and d
    //               are not in the same order as in /usr/include/regex.h
    //
    cflags |=  1;                // use extended regex syntax
    cflags |=  2 * ignore_case;
    cflags |=  4;                // do not match newlines
    // OFF // cflags |= 8;       // only report matches

#if USE_PCRE
    if ((comp_error = pcreposix_regcomp (comp_expression, expression.c_str (), cflags)))
#else
    if ((comp_error = regcomp (comp_expression, expression.c_str (), cflags)))
#endif
    {
        switch (comp_error)
        {
        case REG_BADRPT:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("illegal use of repetition operator (*,+,{) as first character")));
            break;

        case REG_BADBR:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("invalid use of back reference operator")));
            break;

        case REG_EBRACE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("unmatched brace interval operators")));
            break;

        case REG_EBRACK:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("unmatched bracket list operators")));
            break;

        case REG_ERANGE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("invalid use of the range operator")));
            break;

        case REG_ECTYPE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("unknown character class name")));
            break;

// probably still not implemented yet
#ifdef REG_ECOLLATE
        case REG_ECOLLATE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("invalid collating element")));
            break;
#endif

// not defined in POSIX
#ifdef REG_EPAREN
        case REG_EPAREN:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("unmatched parenthesis group operators")));
            break;
#endif
        case REG_ESUBREG:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("invalid back reference to a subexpression")));
            break;

// not defined in POSIX
#ifdef REG_EEND
        case REG_EEND:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("non specific error")));
            break;
#endif
        case REG_EESCAPE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("trailing backslash")));
            break;

        case REG_BADPAT:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("invalid use of pattern operators such as group or list")));
            break;

// not defined in POSIX
#ifdef REG_ESIZE
        case REG_ESIZE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("compiled regular expression requires a pattern buffer larger than 64Kb")));
            break;
#endif
        case REG_ESPACE:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("out of memory")));
            break;

        default:
            throw (Murx::Exceptions::RegExCompilationError
                   (_("compilation of regular expression failed")));
        }
    }
}


void Murx::Filter::print () const
{
    std::cout << _("  testmode: ") << ((testmode)? _("yes") : _("no"))
              << std::endl;

    for (unsigned int i = 0; i < rules.size (); i++)
    {
        std::cout << _("  rule ") << (i+1) << ": "
                  << ((rules[i].only_check_body_lines)? _("body") : _("    "))
                  << ((rules[i].ignore_case)? "" : _(" (case)"))
                  << ((rules[i].negativity)? " <>" : "  =")
                  << ": \"" << rules[i].expression << "\"" << std::endl;
    }

    if (minsize > 0)
        std::cout << _("  min. size: ") << minsize << std::endl;
    if (maxsize > 0)
        std::cout << _("  max. size: ") << maxsize << std::endl;

    std::cout << std::endl;
}


void Murx::Filter::clear ()
{
    // empty rules
    // clear containers
/*
    for (unsigned int i = 0; i < rules.size (); i++)
    {
        delete rules[i];
    }
*/
    rules.clear();

    // reset to default values
    minsize = 0;
    maxsize = 0;

    testmode = false;
}


void Murx::Filter::add_rule (Murx::Rule &rule)
{
    rules.push_back (rule);
}
void Murx::Filter::add_rule (const char* expression, bool ignore_case,
                             bool negativity, bool only_check_body_lines)
{
    if (expression == 0)
    {
        throw (Murx::Exceptions::EmptyPattern (_("empty pattern")));
    }

    std::string expr = expression;
    regex_t     comp_expression;

    compile_expression (expr,ignore_case,&comp_expression);

    // add new filter rule
    Murx::Rule rule;
    rule.expression = expr;
    rule.ignore_case = ignore_case;
    rule.negativity = negativity;
    rule.compiled_expression = comp_expression;
    rule.only_check_body_lines = only_check_body_lines;

    // rule definition is now complete
    rules.push_back (rule);
}
const std::vector<Murx::Rule>& Murx::Filter::get_rules () const
{
    return rules;
}


bool Murx::Filter::get_testmode () const
{
    return testmode;
}
void Murx::Filter::set_testmode (bool yes_or_no)
{
    testmode = yes_or_no;
}


unsigned int Murx::Filter::get_minsize () const
{
    return minsize;
}
void Murx::Filter::set_minsize (unsigned int minsize)
{
    this->minsize = minsize;
}


unsigned int Murx::Filter::get_maxsize () const
{
    return maxsize;
}
void Murx::Filter::set_maxsize (unsigned int maxsize)
{
    this->maxsize = maxsize;
}
