// Filter.hxx - part of the murx project
//
// (c) 2004 by Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifndef MURX_FILTER_HXX
#define MURX_FILTER_HXX

#include <vector>
#include <string>
extern "C"
{
#include <sys/types.h>
#if USE_PCRE
#  include <pcreposix.h>
#else
#  include <regex.h>
#endif
}

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


namespace Murx
{
    class Rule
    {
    public:
        std::string expression;
        bool        ignore_case;
        bool        negativity;
        regex_t     compiled_expression;

        // rule for body lines, do not apply this rule on header lines
        bool        only_check_body_lines;
    };

    // ----------------------------------------

    class Filter
    {
    protected:
        // every rule ist made up of one or more regular expressions
        // concatenated by a logical AND. For consistency there are only
        // set-methods to add all three fields simultaneously to the vectors
        //
        // a rule consists of four fields:
        //
        //   expression, ignore_case, negativity and compiled_expression
        //
        std::vector<Rule>  rules;

        // do not perform any action (approve, delete, move) if set to true
        bool               testmode;

        // minimum size, that the mail must have
        unsigned int       minsize;
        // maximum size, that the mail can have
        unsigned int       maxsize;
        // for both features above: 0 = off

        // compile flags for regcomp() from <regex.h>
        //
        // bits: dcba
        //
        // a : REG_EXTENDED (1 = extended syntax, 0 = basic syntax (deprecated))
        // b : IGNORE_CASE  (1 = yes, 0 = no)
        // c : REG_NEWLINE  (1 = anchors do not match at newline)
        // d : REG_NOSUB    (1 = only report success or failure, no substring)
        //
        // ^--- the manpage of regcomp(3) is buggy because bits c and d
        //      are not in the same order as in /usr/include/regex.h
        //
        // helper function of add_filter (...)
        void compile_expression (std::string expression, bool ignore_case,
                                 regex_t* comp_expression);
  
    public:
        Filter  ();
        Filter  (const char* expression, bool ignore_case,
                 bool negativity, bool only_check_body_lines);
        ~Filter ();

        void                print          () const;
        void                clear          ();

        void                add_rule       (Rule &rule);
        void                add_rule       (const char* expression ,
                                            bool ignore_case,
                                            bool negativity,
                                            bool only_check_body_lines);
        const std::vector<Rule>& get_rules () const;

        bool                get_testmode   () const;
        void                set_testmode   (bool yes_or_no);

        unsigned int        get_minsize    () const;
        void                set_minsize    (unsigned int minsize);

        unsigned int        get_maxsize    () const;
        void                set_maxsize    (unsigned int maxsize);
    };

} /* namespace Murx */

#endif /* MURX_FILTER_HXX */
