// IMAPProtocol.hxx - part of the murx project
//
// (c) 2005 by Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <string>
#include <vector>
#include <sstream>
#include <iostream>
#include <cctype>      // isdigit()

#include "Account.hxx"
#include "Exceptions.hxx"
#include "Header.hxx"
#include "Logger.hxx"
#include "RFC822Parser.hxx"
#include "Socket.hxx"
#include "TCPSocket.hxx"
#if USE_OPENSSL
#include "TLSSocket.hxx"
#endif
#include "Utils.hxx"
#include "i18n.hxx"
#include "IMAPProtocol.hxx"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef logger_logfile
#define logger_logfile Murx::Logger::get_instance(Murx::Logger::Logfile)
#endif

#ifndef logger_netlog
#define logger_netlog Murx::Logger::get_instance(Murx::Logger::Netlog)
#endif

//#define DEBUG_MODE
#undef DEBUG_MODE

#ifdef DEBUG_MODE
# ifdef DEBUG_MSG
#  undef DEBUG_MSG
# endif
# define DEBUG_MSG(msg) std::cerr << msg << std::endl;
#else
# define DEBUG_MSG(msg) ;
#endif


Murx::IMAPProtocol::IMAPProtocol ()
    : sock               (0),
      command            (""),
      str                (""),
      current_message    (0),
      messages_available (0),
      tag_counter        (0),
      do_expunge         (false),
      header_parser      (0)
{
}


Murx::IMAPProtocol::IMAPProtocol (const Murx::Account &account)
    : account            (account),
      sock               (0),
      command            (""),
      str                (""),
      current_message    (0),
      messages_available (0),
      tag_counter        (0),
      do_expunge         (false),
      header_parser      (0)
{
}


Murx::IMAPProtocol::~IMAPProtocol ()
{
    if (sock != 0)
    {
        sock->close();
        delete sock;
    }
}


void Murx::IMAPProtocol::capability()
{
    DEBUG_MSG("DEBUG: capability() started");

    command = get_next_tag() + " CAPABILITY";
    logger_netlog->log_not_console("> " + command);

    sock->write(command + "\r\n");
    
    read_response(response);

    if (! is_request_ok(response))
    {
        // logout
        this->logout();

        // No CAPABILITY warning!
        throw Murx::Exceptions::Protocol
            (_("IMAP server does not support command CAPABILITY"));
    }

    // search for CAPABILITY line in server response
    unsigned int i = find_in_response(response,"* CAPABILITY");

    // found?
    if (i < response.size())
    {
        // i contains index of vector response with capabilities of the server
        str = response[i];
                
        // DEBUG_MSG("DEBUG: CAPABILITY string: " << str);
        
        // >>> CAPABILITY: get all capabilities we use
        unsigned int pos;
        
        // >>> IMAP4rev1
        pos = str.find(" IMAP4rev1",0);
        
        if (pos >= 0 && pos < str.length())
            capabilities.imap4rev1 = true;
        // <<< IMAP4rev1

        // >>> IMAP4
        pos = str.find(" IMAP4",0);
        
        if (pos >= 0 && pos < str.length())
            capabilities.imap4 = true;
        // <<< IMAP4
        
        // >>> STARTTLS
        pos = str.find(" STARTTLS",0);
        
        if (pos >= 0 && pos < str.length())
            capabilities.starttls = true;
        // <<< STARTTLS

        // <<< CAPABILITY
    }
    
    DEBUG_MSG("DEBUG: Capablilites of the IMAP server:");
    DEBUG_MSG("       IMAP4rev1: " << capabilities.imap4rev1);
    DEBUG_MSG("       IMAP4    : " << capabilities.imap4);
    DEBUG_MSG("       STARTTLS : " << capabilities.starttls);

    DEBUG_MSG("DEBUG: capability() finished");
}


void Murx::IMAPProtocol::open_inbox (bool read_only)
{
    command = get_next_tag();

    if (read_only)
    {
        DEBUG_MSG("DEBUG: open_inbox(read-only) started");

        command += " EXAMINE INBOX";
    }
    else
    {
        DEBUG_MSG("DEBUG: open_inbox(read-write) started");

        command += " SELECT INBOX";
    }

    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");
    
    read_response(response);
    
    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }

    DEBUG_MSG("DEBUG: open_inbox() finished");
}


void Murx::IMAPProtocol::connect ()
{
    DEBUG_MSG("DEBUG: connect() started");

    if (account.is_using_tls() && !account.is_using_starttls())
    {    
#if USE_OPENSSL
        sock = new Murx::Net::TLSSocket( account.get_server(),
                                         account.get_port(),
                                         account.get_timeout(),
                                         account.get_fingerprint() );
        sock->connect();
        
        // read server greeting (1 line)
        str.clear();
        read_line(str);
        logger_netlog->log_not_console("< " + str);

        // TODO: check if server refused connection, or preauth, or ...

        // get capabilities from IMAP server
        capability();
#else
        // This code will be only executed if an account requires
        // TLS, but murx was compiled without OpenSSL.

        throw Murx::Exceptions::Protocol
            (_("this version of ")
             + static_cast<std::string> (PACKAGE_NAME)
             + _(" was compiled without SSL support"));
#endif
    }
    else
    {
        sock = new Murx::Net::TCPSocket(account.get_server(),
                                        account.get_port(),
                                        account.get_timeout());
        sock->connect();

        // read server greeting (1 line)
        str.clear();
        read_line(str);
        logger_netlog->log_not_console("< " + str);

        // get capabilities from IMAP server
        capability();

#if USE_OPENSSL
        if (account.is_using_starttls())
        {
            throw Murx::Exceptions::Protocol(_("IMAP protocol error: STARTTLS not implemented yet"));
/*
            if (this->capabilities.starttls)
            {
                command = get_next_tag() + " STARTTLS";
                logger_netlog->log_not_console("> " + command);
                sock->write(command + "\r\n");

                read_response(response);

                if (! is_request_ok(response))
                {
                throw Murx::Exceptions::Protocol
                (_("IMAP protocol error: on sending command \"")
                + command + _("\" server returns an error"));
                }
                else
                {
                    sock = new Murx::Net::TLSSocket
                        ((Murx::Net::TCPSocket *)(sock),
                         account.get_fingerprint());
                }
            }
            else
            {
                // logout and exit with exception
                this->logout();

                // No STARTTLS warning!
                throw Murx::Exceptions::SSL
                    (_("IMAP server does not support command STARTTLS"));
            }
*/
        }
#else
        // This code will be only executed if an account requires
        // TLS, but murx was compiled without OpenSSL.

        throw Murx::Exceptions::Protocol
            (_("this version of ")
             + static_cast<std::string> (PACKAGE_NAME)
             + _(" was compiled without SSL support"));
#endif
    }

    // check for supported IMAP versions
    if (! (capabilities.imap4rev1 || capabilities.imap4))
    {
        // logout and exit with exception
        this->logout();

        throw Murx::Exceptions::Protocol
            ("IMAP version 4 not supported by "
             + static_cast<std::string> (PACKAGE_NAME));
    }

    DEBUG_MSG("DEBUG: connect() finished");
}


void Murx::IMAPProtocol::authenticate ()
{
    DEBUG_MSG("DEBUG: authenticate() started");

    // there are two ways to authenticate to the server:
    //
    // either you send a "LOGIN user password" command or you send an
    // "AUTHENTICATE method" followed by the login data in the format of the
    // specified method (e.g. Kerberos V4)
    //
    // currently we only support the LOGIN method

    command = get_next_tag() + " LOGIN "
        + account.get_user() + " " 
        + account.get_password() + "\r\n";

    logger_netlog->log_not_console("> " + get_current_tag() + " LOGIN "
                       + account.get_user()
                       + " *****");

    sock->write(command);

    read_response(response);

    if (! is_request_ok(response))
    {
        // login failed
        // logout and exit with exception
        this->logout();

        throw Murx::Exceptions::UnknownAccount();
    }

    // open INBOX read-only
    open_inbox(true);
    
    DEBUG_MSG("DEBUG: authenticate() finished");
}


void Murx::IMAPProtocol::login ()
{
    // connect to the IMAP server ...
    this->connect();

    // ... and authenticate
    this->authenticate();
    
    // get count of recent messages
    fetch_message_count();
}


void Murx::IMAPProtocol::logout ()
{
    DEBUG_MSG("DEBUG: logout() started");

    if (do_expunge)
    {
        command = get_next_tag() + " EXPUNGE";
        logger_netlog->log_not_console("> " + command);
        sock->write(command + "\r\n");
        
        read_response(response);
        
        if (! is_request_ok(response))
        {
            throw Murx::Exceptions::Protocol
                (_("IMAP protocol error: on sending command \"")
                 + command + _("\" server returns an error"));
        }
    }

    command = get_next_tag() + " LOGOUT";
    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");

    read_response(response);

    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }
    
    sock->close();

    DEBUG_MSG("DEBUG: logout() finished");
}


int Murx::IMAPProtocol::get_message_count () const
{
    return messages_available;
}


void Murx::IMAPProtocol::fetch_message_count ()
{
    DEBUG_MSG("DEBUG: fetch_message_count() started");

    // index of line in response
    unsigned int line;

    // begin of substring
    unsigned int pos;

    // end of substring
    unsigned int end;

    // get message uids of recent messages
    command = get_next_tag() + " UID SEARCH NEW";

    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");
    
    read_response(response);
    
    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }


    line = find_in_response(response,"* SEARCH ",0);

    DEBUG_MSG("DEBUG: lines of SEARCH resonse: " << response.size());

    // check all returned "* SEARCH ..." lines
    while (line < response.size())
    {
        DEBUG_MSG("DEBUG: Found * SEARCH in line: " << line);

        // length of string "* SEARCH "
        pos = 9;
        end = response[line].length()-1;
        
        // DEBUG_MSG("DEBUG: length = " << end);
        
        // collect all message UIDs in actual line
        while (end < response[line].length())
        {
            // uids are separated by a space
            end = response[line].find(" ",pos+1);
            
            if (end > response[line].length())
                end = response[line].length();
            
            // DEBUG_MSG("DEBUG: pos = " << pos << ", end = " << end);
            
            if (end > pos)
            {
                // get substring of uid, atoi and add to vector
                msg_ids.push_back(atoi(response[line].substr(pos,end).c_str()));

                DEBUG_MSG("DEBUG: Added "
                          << atoi(response[line].substr(pos,end).c_str())
                          << " to msg_ids");
                
                // check next uid
                pos = end;
            }
        }

        // is there another "* SEARCH ..." line?
        line = find_in_response(response,"* SEARCH ",line+1);

        DEBUG_MSG("DEBUG: Next line: " << line);
    }

    DEBUG_MSG("DEBUG: count of new messges is " << msg_ids.size());

    // the message count is the size of the vector with the uids
    messages_available = msg_ids.size();

    DEBUG_MSG("DEBUG: fetch_message_count() finished");
}


bool Murx::IMAPProtocol::get_next_header (Murx::Header &header,
                                          const int body_lines)
{
    DEBUG_MSG("DEBUG: get_next_header() started");

    if (current_message >= messages_available)
    {
        // no more messages
        return false;
    }


    // set the serverside Message ID of the header object
    header.set_id(msg_ids[current_message]);


    // >>> get the message size
    command = get_next_tag() + " UID FETCH "
        + Murx::Utils::int_to_string(msg_ids[current_message])
        + " (FLAGS RFC822.SIZE)";

    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");
    
    read_response(response);
    
    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }

    unsigned int i = find_in_response(response,
                                      " UID "
                                      + Murx::Utils::int_to_string(msg_ids[current_message])
                                      + " RFC822.SIZE ",0);

    if (i < response.size())
    {
        unsigned int begin = response[i].find(" RFC822.SIZE ",0);
        unsigned int end;

        // skip " RFC822.SIZE "
        begin += 13;
                
        for (end = begin; end < response[i].length(); end++)
        {
            if (! isdigit(response[i][end]))
                break;
        }

        // DEBUG_MSG("DEBUG: length = " << response[i].length());
        // DEBUG_MSG("       begin  = " << begin);
        // DEBUG_MSG("         end  = " << end);
        
        header.set_size(atoi(response[i].substr(begin,end).c_str()));
        
        DEBUG_MSG("DEBUG: Set message size to " << header.get_size());
    }

    // <<< get the message size



    // >>> fetch the header

    command = get_next_tag() + " UID FETCH "
        + Murx::Utils::int_to_string(msg_ids[current_message])
        + " (FLAGS RFC822.HEADER)";

    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");
    
    read_response(response);
    
    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }

    // separate header from other stuff
    //
    // an example of the response of the above FETCH command
    //
    // {tag} UID FETCH {uid} (FLAGS RFC822.HEADER)
    // * {id} UID FETCH {uid} (FLAGS () RFC822.HEADER)
    // {... header lines ...}
    // {... empty line ...}    <-- header/body separator
    // )
    // {tag} OK UID FETCH Completed

    // no header (unusual, probably sent wrong uid?)
    if (response.size() < 2)
    {
        DEBUG_MSG("DEBUG: there was no header in return");
        return false;
    }

    // <<< fetch the header



    // >>> parse the header

    // copy header from response to a serialized string
    std::string str;

    // header starts at line 2 and ends two lines before OK response
    for (unsigned int i = 1; i < response.size()-2; i++)
        str += response[i];

    // DEBUG_MSG("DEBUG header (str) = \r\n" << str << ")");
    


    // >>> fetch part of the body

    // amount of octets to fetch from the body

    if (body_lines > 0)
    {
        unsigned int body_octets = OCTETS_PER_LINE * body_lines;
        
        DEBUG_MSG("DEBUG: body_octets results to " << Murx::Utils::int_to_string(body_octets));
        
        // TODO: I don't know, if current_message+1 is always the index
        //       corresponding to the correct UID of the message.
        //
        //       the problem is that PARTIAL does not support UID.
        //
        command = get_next_tag() + " PARTIAL "
            + Murx::Utils::int_to_string(current_message+1)
            + " RFC822.TEXT.PEEK 1 "
            + Murx::Utils::int_to_string(body_octets);
        
        logger_netlog->log_not_console("> " + command);
        sock->write(command + "\r\n");
        
        read_response(response);
        
        if (! is_request_ok(response))
        {

#ifdef DEBUG_MODE

            std::cerr << "DEBUG: IMAP server response following:" << std::endl;
            for (unsigned int i = 0; i < response.size(); i++)
                std::cerr << "DEBUG: " << response[i];

#endif /* DEBUG_MODE */

/*
            throw Murx::Exceptions::Protocol
                (_("IMAP protocol error: on sending command \"")
                 + command + _("\" server returns an error"));
*/

            // this is a non-critical error
            // just log and continue
            std::string logmessage
                = static_cast<std::string> (_("warning: "))
                + _("this IMAP server does not support fetching part of the body");
            logger_logfile->log(logmessage, Murx::Logger::WARNING);

            logmessage
                = static_cast<std::string> (_("warning: "))
                + _("set BODYLINES to zero for this account to get rid of this message");
            logger_logfile->log(logmessage, Murx::Logger::WARNING);
        }
        else
        {
            // append read body lines to header
            for (unsigned int i = 1; i < response.size(); i++)
                str += response[i];
        }
    }

    // <<< fetch part of the body



    // use string as header input stream
    std::istringstream *header_instream = new std::istringstream(str);

    header_parser = new Murx::RFC822Parser(header_instream);
    
    try
    {
        // parse the header
        header_parser->parse (&header);
    }
    catch (Murx::Exceptions::LexicographicalError &e)
    {
        // Parser could not handle header (not rfc822 conformant header?)

        DEBUG_MSG("DEBUG: caught LexicographicalError in header");

        delete header_parser;
        delete header_instream;

        // tell the user what's happened
        std::string logmessage
            = _("message ") + Murx::Utils::int_to_string(current_message)
            + _(" with id = \"") + header.get_message_id()
            + _("\" caused a parser error");
        logger_logfile->log(logmessage, Murx::Logger::DEBUG);

        // don't read the same header again...
        current_message++;

        // handling in Session.cxx: filter_next_mail()
        throw;
    }
    catch (Murx::Exceptions::SyntaxError &e)
    {
        // this exception could but should not occur
        // -> ignore

        DEBUG_MSG("DEBUG: caught SyntaxError in header");
    }
    
    delete header_parser;
    delete header_instream;

    // <<< parse the header


    // don't read the same header again...
    current_message++;

    DEBUG_MSG("DEBUG: get_next_header() finished");

    return true;
}


bool Murx::IMAPProtocol::is_request_ok (const std::vector<std::string>& response)
{
    if (response.size() > 0)
    {
        // compare the last line of the server answer with "{current_tag} OK"
        if (response[response.size()-1].compare
            (0, get_current_tag().length() + 3, get_current_tag() + " OK") == 0)
        {
            return true;
        }
    }

    // answer was either {current_tag} NO or {current_tag} BAD
    return false;
}


unsigned int Murx::IMAPProtocol::find_in_response (const std::vector<std::string>& response,
                                                   const std::string& substring,
                                                   unsigned int       start_pos)
{
    //DEBUG_MSG("DEBUG: Looking for substring \"" << substring << "\"");

    for (unsigned int i = start_pos; i < response.size(); i++)
    {
        //DEBUG_MSG("DEBUG: find_in_response(): response[" << i << "] = " << response[i]);

        if (response[i].find(substring,0) < response[i].length())
            return i;   // return index
    }

    // substring not found
    return response.size();
}


void Murx::IMAPProtocol::read_line (std::string &str)
{
    do
        sock->read(str);
    while (str[str.length()-1] != '\n');
}


void Murx::IMAPProtocol::read_response (std::vector<std::string> &response)
{
    //DEBUG_MSG("DEBUG: read_response() started");

    unsigned int pos;
    std::string  line;
    std::string  pattern = get_current_tag() + " ";

    // read response
    response.clear();
    do
    {
        line.clear();
        read_line(line);

        // check if we read multiple lines
        pos = line.find("\r\n",0);

        // is pos the last line end? (-> single line)
        if (pos < line.length()-2)
        {
            // read multiple lines
            //DEBUG_MSG("DEBUG: read multiple lines");

            // split multiple lines into single lines and add to vector

            // check if the found newline is the last one
            while (pos < line.length()-2)
            {
                //DEBUG_MSG("DEBUG: pos = " << pos);
                //DEBUG_MSG("DEBUG: added line \"" << line.substr(0,pos) << "\"");

                // add line to vector
                response.push_back(line.substr(0,pos+2));
                
                // remove line from multiple lines
                line.erase(0, pos+2);

                // position of next line end
                pos = line.find("\r\n",0);
            }
        }

        // single line or last line of multiple lines
        response.push_back(line);

        // check as long line is not empty (erased by multiple line handler or
        // line begins with "{TAG} " which should be the last line of a server
        // answer)
    } while (line.find(pattern) != 0);
             //&& line.compare(0, pattern.length(), pattern) != 0);
             
    // log all response
    for (unsigned int i = 0; i < response.size(); i++)
        logger_netlog->log_not_console("< " + response[i]);

    //DEBUG_MSG("DEBUG: read_response() finished");
}


void Murx::IMAPProtocol::delete_message (unsigned int id)
{
    DEBUG_MSG("DEBUG: delete_message() started");

    // open INBOX read-write
    open_inbox(false);

    command = get_next_tag() + " UID STORE "
        + Murx::Utils::int_to_string(id)
        + " +FLAGS (\\Deleted)";
    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");

    read_response(response);

    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }

    // set flag for sending an EXPUNGE in logout()
    do_expunge = true;

    DEBUG_MSG("DEBUG: delete_message() finished");
}


void Murx::IMAPProtocol::move_message (unsigned int id,
                                       const std::string& folder)
{
    DEBUG_MSG("DEBUG: move_message() started");

    // check if folder exists in mailbox
    command = get_next_tag() + " EXAMINE " + folder;
    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");

    read_response(response);

    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::FolderDoesNotExist(folder);
    }

    // open INBOX read-write
    open_inbox(false);

    // copy message to folder
    command = get_next_tag() + " UID COPY "
        + Murx::Utils::int_to_string(id) + " " + folder;
    logger_netlog->log_not_console("> " + command);
    sock->write(command + "\r\n");

    read_response(response);

    if (! is_request_ok(response))
    {
        throw Murx::Exceptions::Protocol
            (_("IMAP protocol error: on sending command \"")
             + command + _("\" server returns an error"));
    }

    // delete message from INBOX
    delete_message(id);

    DEBUG_MSG("DEBUG: move_message() finished");
}


std::string Murx::IMAPProtocol::get_current_tag()
{
    return Murx::Utils::int_to_string(tag_counter);
}


std::string Murx::IMAPProtocol::get_next_tag()
{
    return Murx::Utils::int_to_string(++tag_counter);
}
