// IMAPProtocol.hxx - part of the murx project
//
// (c) 2005 by Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifndef MURX_IMAPPROTOCOL_HXX
#define MURX_IMAPPROTOCOL_HXX

#include <string>
#include <vector>
#include <sstream>

#include "Account.hxx"
#include "Header.hxx"
#include "Protocol.hxx"
#include "RFC822Parser.hxx"
#include "Socket.hxx"


namespace Murx
{
    class Capabilities
    {
    public:
        bool imap4rev1;
        bool imap4;
        bool starttls;
        
        Capabilities()
            {
                imap4rev1 = false;
                imap4     = false;
                starttls  = false;
            }
    };   // class Capabilites

    class IMAPProtocol : public Protocol
    {
    private:
        // used for calculation of peeked octets of a message body
        // IMAP does not support POP3-style line orientated TOP 1 5 command
        // We simply assume a line contains 80 characters
        static const unsigned int OCTETS_PER_LINE = 80;

        // account data including all filters
        Account      account;

        // network socket
        Net::Socket  *sock;

        // session states
        //enum SessionState
        //{
        //    Initial          = 0,
        //    NonAuthenticated = 1,
        //    Authenticated    = 2,
        //    Selected         = 3,
        //    Logout           = 4
        //};
        //
        //unsigned int session_state;

        // capabilities of the IMAP server
        Capabilities capabilities;

        // client request (1 line)
        std::string               command;

        // server answer (multiple lines)
        std::vector<std::string>  response;

        std::string  str;
        
        int          current_message;
        int          messages_available;

        // counter for the actually tag (id-prefix for all IMAP commands)
        int          tag_counter;

        // flag used in logout()
        //
        // if set, the EXPUNGE command should be send to server before logout
        // to delete all mails with (\Deleted) flag set.
        bool         do_expunge;

        RFC822Parser* header_parser;
    
        // container for the (unread) message ids
        std::vector<unsigned int> msg_ids;

        std::string  get_current_tag     ();
        std::string  get_next_tag        ();

        void         read_line           (std::string& response);

        void         read_response       (std::vector<std::string>& response);

        void         fetch_message_count (); // throws Protocol

        void         capability          ();               // throws Protocol

        void         open_inbox          (bool read_only); // throws Protocol

        bool         is_request_ok       (const std::vector<std::string>& response);

        unsigned int find_in_response    (const std::vector<std::string>& response,
                                          const std::string& substring,
                                          unsigned int       start_pos = 0);

    public:
        IMAPProtocol ();
        IMAPProtocol (const Account &account);
        ~IMAPProtocol();

        // >>> interface from Protocol.hxx

        //! Creates a connection to the IMAP server defined in account.
        /*! It also initiates encryption via TLS if desired.
         */
        void         connect           (); // throws Protocol, SSL

        //! Sends username and password to the IMAP server.
        void         authenticate      (); // throws UnknownAccount

        //! Wrapper for the sequence connect(); authenticate(); fetch_message_count().
        void         login             (); // throws Protocol, UnknownAccount, SSL

        //! Quits the IMAP session.
        void         logout            (); // throws Protocol

        //! Get the next RFC822 header.
        bool         get_next_header   (Header& header,
                                        const int body_lines = 0);
        //                                    throws Protocol

        //! Get the amount of recent messages on the IMAP server.
        int          get_message_count () const;

        //! Remove the message with ID id on the server.
        void         delete_message    (unsigned int id);

        // <<< interface from Protocol.hxx

        // additional IMAP functionality

        //! Move the message with ID id to a different folder on the server.
        void         move_message      (unsigned int id,
                                        const std::string& folder);
        //                                    throws Protocol, FolderDoesNotExist

    };   // class IMAPProtocol

}   // namespace Murx

#endif /* MURX_IMAPPROTOCOL_HXX */
