// Logger.cxx - part of the murx project
//
// (c) 2004 by Kai Hildebrandt, Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <string>
#include <sstream>
#include <iostream>
#include <ctime>     // time_t

#include "i18n.hxx"
#include "Exceptions.hxx"
#include "Preferences.hxx"
#include "Utils.hxx"
#include "Logger.hxx"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


//
// static private
//

Murx::Logger* Murx::Logger::loggers[COUNT_OF_LOGFILES] = { 0, 0, 0 };

std::string Murx::Logger::month[12] = {
    "Jan", "Feb", "Mar", "Apr",
    "May", "Jun", "Jul", "Aug",
    "Sep", "Oct", "Nov", "Dec"
};


std::string Murx::Logger::get_date ()
{
    std::ostringstream o_stream;
    time_t t = time (0);
    tm* lt = localtime (&t);
    
    o_stream << month[lt->tm_mon] << " ";
    o_stream.width(2); o_stream.fill(' ');
    o_stream << lt->tm_mday << " ";
    o_stream.width(2); o_stream.fill('0');
    o_stream << lt->tm_hour << ":";
    o_stream.width(2); o_stream.fill('0');
    o_stream << lt->tm_min << ":";
    o_stream.width(2); o_stream.fill('0');
    o_stream << lt->tm_sec;
    
    return o_stream.str();
}


//
// static public
//
Murx::Logger* Murx::Logger::get_instance (LogfileIndex index)
{
    unsigned int loglevel = Murx::Preferences::get_instance()->get_loglevel();
    std::string  filename = "";

    // instance already created?
    if (loggers[index] == 0)
    {
        // no -> create new instance of Logger
        //       and open the output filestream for the appropriate
        //       logfile defined in Preferences

        loggers[index] = new Murx::Logger();
        loggers[index]->o_logfile = new std::ofstream();

        switch (index)
        {
        case Logfile:
            // if loglevel == 0 no logfile need to be opened
            // this prevents the creation of an empty logfile
            if (loglevel > 0)
            {
                filename = Murx::Preferences::get_instance()->get_logfile();

                if (filename.length() > 0)
                    if (! loggers[index]->open_logfile(filename))
                        throw Murx::Exceptions::CouldNotOpenFile(filename);
            }
            break;
            
        case Headerfile:
            // if special bit SHOW_HEADERS is not set no headerfile
            // need to be opened
            // this prevents the creation of an empty headerfile
            if ((loglevel & Murx::Logger::SHOW_HEADERS))
            {
                filename = Murx::Preferences::get_instance()->get_headerfile();

                if (filename.length() > 0)
                    if (! loggers[index]->open_logfile(filename))
                        throw Murx::Exceptions::CouldNotOpenFile(filename);
            }
            break;
            
        case Netlog:
            filename = Murx::Preferences::get_instance()->get_netlog();

            if (filename.length() > 0)
                if (! loggers[index]->open_logfile(filename))
                    throw Murx::Exceptions::CouldNotOpenFile(filename);

            break;
        }
    }

    // yes -> return instance of Logger
    //        (file should be open)
    return loggers[index];
}


void Murx::Logger::close (LogfileIndex index)
{
    if (loggers[index] != 0)
    {
        if (loggers[index]->o_logfile != 0)
        {
            // close logfile
            loggers[index]->o_logfile->close();
            
            // free memory
            delete loggers[index]->o_logfile;
        }
    
        delete loggers[index];
        loggers[index] = 0;
    }
}


void Murx::Logger::destroy_all ()
{
    for (unsigned int i = 0; i < COUNT_OF_LOGFILES; i++)
    {
        Murx::Logger::close (LogfileIndex(i));
    }
}


//
// private
//

bool Murx::Logger::open_logfile (const std::string filename)
{
    if (filename.length())
    {
        o_logfile->open(filename.c_str(), std::ios_base::app);

        if (o_logfile->is_open())
            return true;
    }

    return false;
}


//
// public
//

void Murx::Logger::log (std::string message, unsigned int level)
{
    Murx::Utils::trim(message);

    if (o_logfile != 0 && o_logfile->is_open())
    {
        unsigned int loglevel = Murx::Preferences::get_instance()->get_loglevel();
        unsigned int special_bits;

        // isolate the special bits
        // (bit 3 and above have special meanings)
        special_bits = level & ~7;

        // remove special bits set in level
        level  = level & 7;
        
        // if level is greater equal loglevel we print a message and log to file
        if (level > 0 && (loglevel & 7) >= level)
        {
            // Log errors and warnings to stderr.
            //
            // This is useful for automatic filtering via return value because
            // you don't know what happened on return value 0
            // (no mail or error?).
            //
            // With this you can do something like
            //
            //     $ murx > murx.log 2> murx.err
            //
            if (level <= WARNING)
                std::cerr << get_date() << " " << message << std::endl;
            else
                std::cout << get_date() << " " << message << std::endl;
            
            *o_logfile << get_date() << " " << message << std::endl;
            o_logfile->flush();
        }

        // check if a special bit is set
        if (special_bits > 0)
        {
            // is the special bit set in loglevel?
            if ((loglevel & special_bits) == special_bits)
            {
                // yes -> log to file and to console)
                std::cout  << get_date() << " " << message << std::endl;

                *o_logfile << get_date() << " " << message << std::endl;
                o_logfile->flush();
            }
        }
    }
}


void Murx::Logger::log_not_console (std::string message)
{
    Murx::Utils::trim(message);

    if (o_logfile != 0 && o_logfile->is_open())
    {
        *o_logfile << get_date() << message << std::endl;
        o_logfile->flush();
    }
}


void Murx::Logger::log_header (std::string message)
{
    Murx::Utils::trim(message);

    if (o_logfile != 0 && o_logfile->is_open())
    {
        *o_logfile << message << std::endl;
        o_logfile->flush();
    }
}
