// Logger.hxx - part of the murx project
//
// (c) 2004 by Kai Hildebrandt, Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifndef MURX_LOGGER_HXX
#define MURX_LOGGER_HXX

#include <string>
#include <fstream>

namespace Murx
{
    class Logger
    {
        // --------------
        // static members
        // --------------

    private:
        static const unsigned int COUNT_OF_LOGFILES = 3;
        static Logger *loggers[COUNT_OF_LOGFILES];

        static std::string  month[12];
        static std::string  get_date ();
        
    public:
        //! Index for the Logger objects
        enum LogfileIndex
        {
            Logfile    = 0,
            Headerfile = 1,
            Netlog     = 2
        };

        //! Get the appropriate Logger instance
        static Logger* get_instance (LogfileIndex index);
        //                           throws CouldNotOpenFile

        //! Close a specific logfile
        static void          close (LogfileIndex index);

        //! Close open logfiles and destroy all existing instances of Logger
        static void          destroy_all ();

        // definition for verbosity (bitmap):
        //
        // 0000 0000
        //       ^^^ reserved (0-7)
        //
        //   0 - disable logging
        //   1 - errors
        //   2 - warnings, deleted messages
        //   3 - notice, program started (incl. used config file) and finished
        //         messages
        //   4 - info, account name, messages left on account, approved messages
        //   5 - debug
        //   6 - not used (development purposes)
        //   7 - not used (development purposes)
        //
        //   8 - show headers (in a separate headerfile)
        //  16 - show matches
        //  32 - show matches even if the filter did not match
        //  64 - show header line with match
        //

        // --- loglevels ---

        //! Loglevel for error messages
        static const unsigned int ERROR   = 1;

        //! Loglevel for warning messages
        static const unsigned int WARNING = 2;

        //! Loglevel for notice messages
        static const unsigned int NOTICE  = 3;

        //! Loglevel for info messages
        static const unsigned int INFO    = 4;

        //! Loglevel for debug messages
        static const unsigned int DEBUG   = 5;

        // static const unsigned int RESERVED = 6;
        // static const unsigned int RESERVED = 7;

        // special fields (bitmap)

        //! Special bit for logging all complete message headers
        /*! If this special bit is set, log complete headers of all messages
         *  to the headerfile.
         */
        static const unsigned int SHOW_HEADERS = 8;

        //! Special bit for logging matched patterns
        /*! If this special bit is set, log each matched pattern and
         *  matched substring of all applied filters to logfile and
         *  console.
         */
        static const unsigned int SHOW_MATCHES = 16;

        //! Special bit for logging also partly matched patterns
        /*! If this special bit is set, log each matched pattern and
         *  matched substring to logfile and console even if a filter
         *  did not match in every pattern and therewith was not applied.
         */
        static const unsigned int SHOW_ALL_MATCHES = 32;

        //! Special bit for logging complete lines with match
        /*! If this special bit is set, log lines where a pattern matched
         *  to the logfile and console.
         */
        static const unsigned int SHOW_LINE_WITH_MATCH = 64;

        
        // --------------
        // object members
        // --------------

    private:
        // new instances of Logger are created by static method
        // Logger::get_instance(LogfileIndex)
        Logger () {};
        ~Logger() {};

        // open a logfile
        bool           open_logfile    (const std::string filename);

        // output stream for logfile
        std::ofstream* o_logfile;
        
    public:
        //! Log a message if level is greater than loglevel or special bit is set
        void           log (std::string message, unsigned int level = 0);

        //! Log a message to file in any case with date but without console output
        void           log_not_console (std::string message);

        //! Log a message to file in any case without date and without console output
        void           log_header (std::string message);
    };
}

#endif /* MURX_LOGGER_HXX */
