// POP3Protocol.hxx - part of the murx project
//
// (c) 2004 by Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <iostream>
#include <string>
#include <sstream>

#include "Account.hxx"
#include "Exceptions.hxx"
#include "Header.hxx"
#include "Logger.hxx"
#include "RFC822Parser.hxx"
#include "Socket.hxx"
#include "TCPSocket.hxx"
#if USE_OPENSSL
#include "TLSSocket.hxx"
#endif
#include "Utils.hxx"
#include "i18n.hxx"
#include "POP3Protocol.hxx"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef logger_logfile
#define logger_logfile Murx::Logger::get_instance(Murx::Logger::Logfile)
#endif

#ifndef logger_netlog
#define logger_netlog Murx::Logger::get_instance(Murx::Logger::Netlog)
#endif


Murx::POP3Protocol::POP3Protocol ()
    : sock               (0),
      str                (""),
      messages_available (0),
      current_message    (0),
      header_parser      (0)
{
}


Murx::POP3Protocol::POP3Protocol (const Murx::Account &account)
    : account            (account),
      sock               (0),
      str                (""),
      messages_available (0),
      current_message    (1),
      header_parser      (0)
{
}


Murx::POP3Protocol::~POP3Protocol ()
{
    if (sock != 0)
    {
        sock->close();
        delete sock;
    }
}


void Murx::POP3Protocol::connect ()
{
    str.clear();
    
    if (account.is_using_tls())
    {    
#if USE_OPENSSL
        if (account.is_using_starttls())
        {
            throw Murx::Exceptions::Protocol(_("POP3 protocol error: STLS not implemented yet"));
/*
            sock = new Murx::Net::TCPSocket(account.get_server(),
                                            account.get_port(),
                                            account.get_timeout());
            sock->connect();

            read_line(str);
            logger_netlog->log_not_console("< " + str);
            
            if (!is_request_ok(str))
            {
                throw Murx::Exceptions::Protocol(_("could not connect to server"));
            }

            logger_netlog->log_not_console("> STLS");
            sock->write("STLS");

            read_line(str);
            logger_netlog->log_not_console("< " + str);

            if (!is_request_ok(str))
            {
                std::string msg = _("POP3 protocol error:");
                msg += _(" STLS failed");
                throw Murx::Exceptions::Protocol(msg);
            }

            sock = new Murx::Net::TLSSocket((Murx::Net::TCPSocket *)(sock),
                                             account.get_fingerprint());
 */
        }
        else
        {
            sock = new Murx::Net::TLSSocket(account.get_server(),
                                            account.get_port(),
                                            account.get_timeout(),
                                            account.get_fingerprint());
            sock->connect();
        }
#else
        // Murx was compiled without OpenSSL headers

        throw Murx::Exceptions::Protocol
            (_("this version of ")
             + static_cast<std::string> (PACKAGE_NAME)
             + _(" was compiled without SSL support"));
#endif
    }
    else
    {
        sock = new Murx::Net::TCPSocket(account.get_server(),
                                        account.get_port(),
                                        account.get_timeout());
        sock->connect();
    }

    read_line(str);
    logger_netlog->log_not_console("< " + str);

    if (!is_request_ok(str))
    {
        throw Murx::Exceptions::Protocol(_("could not connect to server"));
    }
}


void Murx::POP3Protocol::authenticate ()
{
    str.clear();
    
    logger_netlog->log_not_console("> USER " + account.get_user());
    sock->write("USER " + account.get_user() + "\r\n");

    read_line(str);
    logger_netlog->log_not_console("< " + str);

    if (!is_request_ok(str))
    {
        logger_netlog->log_not_console("> QUIT");
        sock->write("QUIT\r\n");
        throw Murx::Exceptions::UnknownAccount();
    }

    str.clear();
    
    logger_netlog->log_not_console("> PASS *****");
    sock->write("PASS " + account.get_password() + "\r\n");

    read_line(str);
    logger_netlog->log_not_console("< " + str);

    if (!is_request_ok(str))
    {
        logger_netlog->log_not_console("> QUIT");
        sock->write("QUIT\r\n");
        throw Murx::Exceptions::LoginError(str);
    }
}


void Murx::POP3Protocol::login ()
{
    connect();
    authenticate();
    fetch_message_count();
}


void Murx::POP3Protocol::logout ()
{
    str.clear();
    
    logger_netlog->log_not_console("> QUIT");
    sock->write("QUIT\r\n");

    read_line(str);
    logger_netlog->log_not_console("< " + str);

    if (!is_request_ok(str))
    {
        throw Murx::Exceptions::Protocol(_("POP3 protocol error"));
    }
    
    sock->close();
}


int Murx::POP3Protocol::get_message_count () const
{
    return messages_available;
}


void Murx::POP3Protocol::fetch_message_count ()
{
    str.clear();
    
    logger_netlog->log_not_console("> STAT");
    sock->write("STAT\r\n");
    
    read_line(str);
    logger_netlog->log_not_console("< " + str);
    
    if (!is_request_ok(str))
    {
        std::string msg = _("POP3 protocol error:");
        msg += _(" STAT failed");
        throw Murx::Exceptions::Protocol(msg);
    }
    
    std::string::size_type start_pos = 0;
    std::string::size_type end_pos   = 0;
    
    start_pos = str.find(' ', 0) + 1;
    end_pos   = str.find(' ', start_pos);
    
    messages_available = atoi(str.substr(start_pos, end_pos - start_pos).c_str());
}


bool Murx::POP3Protocol::get_next_header (Murx::Header &header,
                                          const int body_lines)
{
    if (current_message > messages_available)
        return false;
    

    // set the serverside messageID of the header object
    header.set_id(current_message);
    
    

    // >>> get the message size

    str.clear();
    std::ostringstream o_stream_list;
    o_stream_list << "LIST " << current_message << "\r\n";
    logger_netlog->log_not_console("> " + o_stream_list.str());
    sock->write(o_stream_list.str());
    
    read_line(str);
    logger_netlog->log_not_console("< " + str);
    if (!is_request_ok(str))
    {
        std::string msg = _("POP3 protocol error:");
        msg += _(" LIST failed");
        throw Murx::Exceptions::Protocol(msg);
    }
    
    std::string::size_type start_pos = 0;
    std::string::size_type end_pos = 0;
    
    // find the first occurence of ' ' BEHIND the message ID ("+OK 4 732")
    start_pos = str.find(' ', 4) + 1;
    end_pos = str.find('\r', start_pos);
    
    header.set_size(static_cast<unsigned int>
                     (atoi
                      (str.substr(start_pos, end_pos - start_pos).c_str())));

    // <<< get the message size
    


    // >>> fetch the header

    std::ostringstream o_stream;
    o_stream << "TOP " << current_message << " " << body_lines << "\r\n";
    logger_netlog->log_not_console("> " + o_stream.str());
    sock->write(o_stream.str());
    
    str.erase();
    do
    {
        sock->read(str);
    }
    while ( !( str[str.length()-5]    == '\r'
               && str[str.length()-4] == '\n'
               && str[str.length()-3] == '.'
               && str[str.length()-2] == '\r'
               && str[str.length()-1] == '\n'));

    logger_netlog->log_not_console("< " + str);

    if (!is_request_ok(str))
    {
        std::string msg = _("POP3 protocol error:");
        msg += _(" TOP failed");
        throw Murx::Exceptions::Protocol(msg);
    }
    
    // we have to delete the leading line containing the "+OK ..."
    str.erase(0, str.find('\n', 0) + 1);
    
    // we have to delete the terminating ".CRLF" sequence
    str.erase(str.length()-3, str.length());
    
    // <<< fetch the header


    // >>> parse the header

    std::istringstream *header_instream = new std::istringstream(str);
    header_parser = new Murx::RFC822Parser(header_instream);
    

    try
    {
        header_parser->parse (&header);
    }
    catch (Murx::Exceptions::LexicographicalError &e)
    {
        // Parser could not handle header (not rfc822 conformant header?)

        delete header_parser;
        delete header_instream;

        // tell the user what's happened
        std::string logmessage
            = _("message ") + Murx::Utils::int_to_string (current_message)
            + _(" with id = \"") + header.get_message_id()
            + _("\" caused a parser error");
        logger_logfile->log(logmessage, Murx::Logger::DEBUG);

        // don't read the same header again...
        current_message++;

        // handling in Session.cxx: filter_next_mail()
        throw;
    }
    catch (Murx::Exceptions::SyntaxError &e)
    {
        // this exception theoretically could but should not occur
    }
    
    delete header_parser;
    delete header_instream;
    
    // <<< parse the header
    
    

    // don't read the same header again...
    current_message++;

    return true;
}


bool Murx::POP3Protocol::is_request_ok (std::string server_answer)
{
    if (server_answer.compare(0, 3, "+OK") != 0)
        return false;
    else
        return true;
}


void Murx::POP3Protocol::read_line (std::string &str)
{
    do
        sock->read(str);
    while (str[str.length()-1] != '\n');
}


void Murx::POP3Protocol::delete_message (unsigned int id)
{
    std::ostringstream o_stream;
    o_stream << "DELE " << id << "\r\n";
    logger_netlog->log_not_console("> " + o_stream.str());
    
    sock->write(o_stream.str());
    
    str.erase();
    read_line(str);
    logger_netlog->log_not_console("< " + str);
    
    if (!is_request_ok(str))
    {
        throw Murx::Exceptions::Protocol(_("could not delete message"));
    }
}
