// Preferences.cxx - part of the murx project
//
// (c) 2004 by Kai Hildebrandt, Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <string>
#include <vector>
#include <iostream>
#include "i18n.hxx"
#include "Account.hxx"
#include "Filter.hxx"
#include "Score.hxx"
#include "MoveTo.hxx"
#include "Exceptions.hxx"
#include "Utils.hxx"
#include "Preferences.hxx"


// initialization of the static members

// Preferences is a singleton object and
// will be only allocated once. This variable
// stores the pointer to the object.
//
Murx::Preferences* Murx::Preferences::preferences = 0;


// get a pointer to the preferences object
Murx::Preferences* Murx::Preferences::get_instance ()
{
    if (preferences == 0)
        preferences = new Murx::Preferences ();
    
    return preferences;
}


Murx::Preferences::Preferences ()
    : loglevel            (3),
      normalize_subject   (0),
      ignore_case         (0),
      testmode            (0),
      body_lines          (0),
      delete_duplicates   (0),
      return_remaining    (0),
      highscore         (100),
      global_timeout     (60),
      maxsize_allow       (0), 
      maxsize_deny        (0),
      maxlength           (0)
{
//    std::cerr << "new Preferences created" << std::endl;

    nonconformant_action = "";
    nonconformant_folder = "";
}


Murx::Preferences::~Preferences ()
{
//    std::cerr << "Preferences destroyed" << std::endl;
}


void Murx::Preferences::destroy ()
{
    // delete the stored object
    // no further usage of "get_instance()" possible!
    if (preferences != 0)
    {
        delete preferences;
        preferences = 0;
    }
}


void Murx::Preferences::print () const
{
    // >>> separate loglevel and special bits
    unsigned int level = loglevel &  7;
    unsigned int sbits = loglevel & ~7;

    std::cout << "loglevel          : " << level;

    for (unsigned int i=8; i<128; i=i*2)
    {
        if ((sbits & i) == i)
            std::cout << " + " << i;
    }
    std::cout << std::endl;
    // <<< separate loglevel and special bits

    std::cout << "logfile           : " << logfile << std::endl;
    std::cout << "headerfile        : " << headerfile << std::endl;
    std::cout << "netlog            : " << netlog << std::endl;
    std::cout << "non-conf. action  : " << nonconformant_action << std::endl;
    std::cout << "non-conf. folder  : " << nonconformant_folder << std::endl;
    std::cout << "normalize_subject : " << ((normalize_subject)? "yes" : "no") << std::endl;
    std::cout << "ignore_case       : " << ((ignore_case)? "yes" : "no")       << std::endl;
    std::cout << "testmode          : " << ((testmode)? "yes" : "no")          << std::endl;
    std::cout << "bodylines         : " << body_lines << std::endl;
    std::cout << "delete_duplicates : " << ((delete_duplicates)? "yes" : "no") << std::endl;
    std::cout << "return_remaining  : " << ((return_remaining)? "yes" : "no")  << std::endl;
    std::cout << "highscore         : " << highscore << std::endl;
    std::cout << "global_timeout    : " << global_timeout << std::endl;
    std::cout << "maxsize_allow     : " << maxsize_allow << std::endl;
    std::cout << "maxsize_deny      : " << maxsize_deny << std::endl;
    std::cout << "maxlength         : " << maxlength << std::endl;
}


void Murx::Preferences::add_account (Murx::Account &account)
{
    accounts.push_back (account);
}
const std::vector<Murx::Account>& Murx::Preferences::get_accounts () const
{
    return accounts;
}
Murx::Account& Murx::Preferences::get_account (std::string name,
                                               unsigned int line)
{
    // search for an account with given name

    for (unsigned int i = 0; i < accounts.size (); i++ )
    {
        if (accounts[i].get_name () == name)
        {
            return accounts[i];
        }
    }

    if (line == 0)
    {
        throw (Murx::Exceptions::AccountNotFound (_("account \"") + name
                                                  + _("\" not found")));
    }
    else
    {
        throw (Murx::Exceptions::AccountNotFound (_("account \"") + name
                                                  + _("\" not found near line ")
                                                  + Murx::Utils::int_to_string (line)));
    }
}


void Murx::Preferences::set_headerfile (std::string headerfile)
{
    this->headerfile = headerfile;
}
std::string Murx::Preferences::get_headerfile () const
{
    return headerfile;
}


void Murx::Preferences::set_logfile (std::string logfile)
{
    this->logfile = logfile;
}
std::string Murx::Preferences::get_logfile () const
{
    return logfile;
}


void Murx::Preferences::set_netlog (std::string netlog)
{
    this->netlog = netlog;
}
std::string Murx::Preferences::get_netlog () const
{
    return netlog;
}


void Murx::Preferences::set_loglevel (unsigned int loglevel)
{
    this->loglevel = loglevel;
}
unsigned int Murx::Preferences::get_loglevel () const
{
    return loglevel;
}


void Murx::Preferences::set_nonconformant_action (std::string action)
{
    nonconformant_action = action;
}
std::string Murx::Preferences::get_nonconformant_action () const
{
    return nonconformant_action;
}


void Murx::Preferences::set_nonconformant_folder (std::string folder)
{
    nonconformant_folder = folder;
}
std::string Murx::Preferences::get_nonconformant_folder () const
{
    return nonconformant_folder;
}


void Murx::Preferences::set_normalize_subject (bool yes_or_no)
{
    normalize_subject = yes_or_no;
}
bool Murx::Preferences::get_normalize_subject () const
{
    return normalize_subject;
}


void Murx::Preferences::set_ignore_case (bool yes_or_no)
{
    ignore_case = yes_or_no;
}
bool Murx::Preferences::get_ignore_case () const
{
    return ignore_case;
}


void Murx::Preferences::set_testmode (bool yes_or_no)
{
    testmode = yes_or_no;
}
bool Murx::Preferences::get_testmode () const
{
    return testmode;
}


void Murx::Preferences::set_body_lines (unsigned int body_lines)
{
    this->body_lines = body_lines;
}
unsigned int Murx::Preferences::get_body_lines () const
{
    return body_lines;
}


void Murx::Preferences::set_delete_duplicates (bool yes_or_no)
{
    delete_duplicates = yes_or_no;
}
bool Murx::Preferences::get_delete_duplicates () const
{
    return delete_duplicates;
}


void Murx::Preferences::set_return_remaining (bool yes_or_no)
{
    return_remaining = yes_or_no;
}
bool Murx::Preferences::get_return_remaining () const
{
    return return_remaining;
}


void Murx::Preferences::set_highscore (int highscore)
{
    this->highscore = highscore;
}
int Murx::Preferences::get_highscore () const
{
    return highscore;
}


void Murx::Preferences::set_global_timeout (unsigned int global_timeout)
{
    this->global_timeout = global_timeout;
}
unsigned int Murx::Preferences::get_global_timeout () const
{
    return global_timeout;
}


void Murx::Preferences::set_maxsize_allow (unsigned int maxsize_allow)
{
    this->maxsize_allow = maxsize_allow;
}
unsigned int Murx::Preferences::get_maxsize_allow () const
{
    return maxsize_allow;
}


void Murx::Preferences::set_maxsize_deny (unsigned int maxsize_deny)
{
    this->maxsize_deny = maxsize_deny;
}
unsigned int Murx::Preferences::get_maxsize_deny () const
{
    return maxsize_deny;
}


void Murx::Preferences::set_maxlength (unsigned int maxlength)
{
    this->maxlength = maxlength;
}
unsigned int Murx::Preferences::get_maxlength () const
{
    return maxlength;
}
