// Session.cxx - part of the murx project
//
// (c) 2004 by Thomas Stauss, Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <string>
#include <vector>
#include <istream>
#include <fstream>

extern "C"
{
#include <sys/types.h>
#if USE_PCRE
#  include <pcreposix.h>
#else
#  include <regex.h>
#endif
}

#include "Session.hxx"

#include "Header.hxx"
#include "Protocol.hxx"
#include "POP3Protocol.hxx"
#include "IMAPProtocol.hxx"
#include "Account.hxx"
#include "Preferences.hxx"
#include "Match.hxx"
#include "MatchData.hxx"
#include "RCParser.hxx"
#include "RFC822Parser.hxx"
#include "Filter.hxx"
#include "MoveTo.hxx"
#include "Score.hxx"
#include "Utils.hxx"
#include "i18n.hxx"


// >>> STATUS FLAGS
const int Murx::Session::ACCOUNT_FILTERING_FINISHED = -1;
const int Murx::Session::NORULE                     =  0;
const int Murx::Session::ALLOW                      =  1;
const int Murx::Session::DELETE_MAXSIZE_ALLOW       =  2;
const int Murx::Session::DELETE_DENYRULE            =  3;
const int Murx::Session::DELETE_MAXSIZE_DENY        =  4;
const int Murx::Session::DELETE_DUPLICATE           =  5;
const int Murx::Session::DELETE_NONCONFORMANT       =  6;
const int Murx::Session::DELETE_MAXLENGTH_EXCEEDED  =  7;
const int Murx::Session::MOVETO                     =  8;
const int Murx::Session::MOVETO_NONCONFORMANT       =  9;
const int Murx::Session::SCORE_DELETE               = 10;
// <<< STATUS FLAGS


Murx::Session::Session (std::string rcfile)
    // first creation of the Preferences object
    : preferences             (*Murx::Preferences::get_instance()),
      want_next_account       (true),
//      account_count           (0),
      current_account         (0),
//      filter_count            (0),
//      rule_count              (0),
      remaining_message_count (0),
      first_run               (true),
      only_size_not_match     (false)
{
    

    // >>> open and parse rcfile

    std::filebuf i_file;
    
    if (rcfile.empty())
    {
        rcfile = getenv("HOME");
        rcfile += "/.murxrc";
    }
    
    i_file.open (rcfile.c_str(), std::ios::in);
    
    if (! i_file.is_open())
        throw Murx::Exceptions::IOError (_("could not open rcfile ") + rcfile);
    
    std::istream i_stream (&i_file);
    Murx::RCParser rcParser (&i_stream);
    rcParser.parse (&preferences);
    i_file.close();

    // <<< open and parse rcfile
}


Murx::Session::~Session ()
{
    Murx::Preferences::destroy();
    
    for (unsigned int i = 0; i < protocol.size(); i++)
    {
        delete protocol[i];
    }
}


void Murx::Session::account_back ()
{
    if (! first_run && current_account > 0)
        current_account--;
}


const Murx::Account& Murx::Session::get_current_account () const
{
    return preferences.get_accounts().at(current_account);
}


std::string Murx::Session::get_current_accountname () const
{
    return preferences.get_accounts().at(current_account).get_name();
}


std::string Murx::Session::get_next_accountname ()
{
    if ((current_account + 1) == preferences.get_accounts().size())
        return "";
    else
        return preferences.get_accounts().at(current_account + 1).get_name();
}


int Murx::Session::get_current_message_count () const
{
    return protocol[protocol.size() - 1]->get_message_count();
}


unsigned int Murx::Session::get_remaining_message_count () const
{
    return remaining_message_count;
}



bool Murx::Session::filter_next_mail (Murx::Header &header)
{
    // default return value
    header.set_action(Murx::Session::ACCOUNT_FILTERING_FINISHED);

    // check if we already have a connection to a server
    // and there are mails left on this server
    if (want_next_account)
    {
        // don't increase current_account on first run
        if (first_run)
            first_run = false;
        else
            current_account++;
        
        // if there are no more accounts to check, we can leave
        if (current_account == preferences.get_accounts().size())
            return false;
        
        
        // create a new protocol object for current account
        if (preferences.get_accounts().at(current_account).get_protocol()
            == "pop3")
        {
            protocol.push_back
                (new Murx::POP3Protocol(
                    preferences.get_accounts().at(current_account)));
        }
        else if (preferences.get_accounts().at(current_account).get_protocol()
                 == "imap")
        {
            protocol.push_back
                (new Murx::IMAPProtocol(
                    preferences.get_accounts().at(current_account)));
        }
        
        // login to server
        try
        {
            protocol[protocol.size() - 1]->login();
            
            want_next_account = false;

            // get the total message count of this server
            remaining_message_count = static_cast<unsigned int>
                (protocol[protocol.size() - 1]->get_message_count());
        }
        catch (Murx::Exceptions::Network &e)
        {
            delete protocol[protocol.size() - 1];
            protocol.erase (protocol.end() - 1);
            throw;
        }
        catch (Murx::Exceptions::Protocol &e)
        {
            delete protocol[protocol.size() - 1];
            protocol.erase (protocol.end() - 1);
            throw;
        }
    }
    
    try
    {
        // try to fetch a mail header from server
        // if there are no more mails left on this server: next account
        //
        // NOTE:
        //
        // the action is still set to ACCOUNT_FILTERING_FINISHED here
        //
        if (protocol[protocol.size() - 1]->get_next_header
            (header, preferences.get_body_lines()) == false)
        {
            // there are no more mails left on this server
            want_next_account = true;

            return true;
        }
    }
    catch (Murx::Exceptions::LexicographicalError &e)
    {
        // we don't want to match any rules with such malformed mails
        // because the header may be not intact

        header.set_action(Murx::Session::NORULE);
        return true;
    }
    catch (Murx::Exceptions::Network &e)
    {
        delete protocol[protocol.size() - 1];
        protocol.erase (protocol.end() - 1);
        
        want_next_account = true;
        
        throw;
    }
    catch (Murx::Exceptions::Protocol &e)
    {
        delete protocol[protocol.size() - 1];
        protocol.erase (protocol.end() - 1);
        
        want_next_account = true;
        
        throw;
    }
    
    
    // set the current account-ID in the header
    //
    // HINT: Why we do not use current_account here?
    //
    // it will lead to a STL exception because of a bad vector index
    // in delete_mail(account,id). the variable current_account only
    // counts from 0 to preferences.get_accounts().size() but does not
    // point on the actually used connection -> protocol.size()-1.
    //
    header.set_account(protocol.size() - 1);
    
    
    // we fetched a header
    // now we have to check all rules in the given sequence:
    //
    // - check for duplicate messages
    //
    // - check all ALLOW rules, if an ALLOW-rule matched, check MAXSIZE_ALLOW
    //
    // - check all DENY rules
    //
    // - check all MOVETO rules
    //
    // - if a DENY or MOVETO filter matched completely except a SIZE rule,
    //   check MAXSIZE_DENY
    //
    // - check all SCORE rules and compare with HIGHSCORE
    
    if (check_conformance(header))
        return true;

    if (check_maxlength(header))
        return true;

    if (check_duplicates(header))
        return true;

    if (check_allows(header))
        return true;

    only_size_not_match = false;

    if (check_movetos(header))
        return true;

    if (check_denies(header))
        return true;

    if (! only_size_not_match)
        if (check_maxsize(header))
            return true;

    if (check_scores(header))
        return true;
    
    
    
    // no rule matched
    header.set_action(Murx::Session::NORULE);
    
    return true;
}



//
// >>> NONCONFORMANT
//
bool Murx::Session::check_conformance (Murx::Header &header)
{
    if (!header.is_rfc2822_conformant())
    {
        if (preferences.get_nonconformant_action().length() != 0)
        {
            if (preferences.get_nonconformant_action() == "DENY")
            {
                // delete message
                header.set_action(Murx::Session::DELETE_NONCONFORMANT);
                
                // check for testmode
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // decrease message count of this account
                    remaining_message_count--;
                }
                
                return true;
            }
            else if (preferences.get_nonconformant_action() == "MOVETO")
            {
                if (preferences.get_nonconformant_folder().length() == 0)
                {
                    // no folder given
                    header.set_action(Murx::Session::NORULE);
                    return true;
                }
                else
                {
                    // move message
                    header.set_action(Murx::Session::MOVETO_NONCONFORMANT);
                    return true;
                }
            }
        }
    }

    return false;
}
//
// >>> NONCONFORMANT
//



//
// >>> MAXLENGTH
//
bool Murx::Session::check_maxlength (Murx::Header &header)
{
    if (preferences.get_maxlength() > 0)
    {
        if (header.get_maxlength() > preferences.get_maxlength())
        {
            header.set_action(Murx::Session::DELETE_MAXLENGTH_EXCEEDED);
            return true;
        }
    }

    return false;
}
//
// <<< MAXLENGTH
//



//
// >>> DUPLICATE MESSAGES
//
bool Murx::Session::check_duplicates (Murx::Header &header)
{
    // should we check for duplicate message ids?
    if (! preferences.get_delete_duplicates())
        return false;

    // check all existing message ids against new one
    for (unsigned int i = 0; i < message_ids.size(); i++)
    {
        if (header.get_message_id() == message_ids[i])
        {
            // found duplicate
            header.set_action(Murx::Session::DELETE_DUPLICATE);

            // decrease message count of this account
            remaining_message_count--;

            return true;
        }
    }
    // not duplicate
    // add new message id to container
    message_ids.push_back(header.get_message_id());

    return false;
}
//
// <<< DUPLICATE MESSAGES
//



//
// >>> ALLOWS
//
bool Murx::Session::check_allows (Murx::Header &header)
{
    // how many allow filters are available?
    const unsigned int filter_count =
        preferences.get_accounts().at(current_account).get_allows().size();
    
    // >>> check all filters of the account
    for (unsigned int current_filter = 0;
         current_filter < filter_count;
         current_filter++)
    {
        // create new vector for temporary saving of the matches
        Murx::Match tmp_match_list;
        
        // get next filter object
        const Murx::Filter& filter =
            preferences.get_accounts().at(current_account).get_allows().at(current_filter);
        
        bool filter_matched = true;


        // check the local sizes

        // maximum size
        if (filter.get_maxsize() != 0
            && (header.get_size() >= filter.get_maxsize()))
            continue;

        // minimal size
        if (filter.get_minsize() != 0
            && (header.get_size() <= filter.get_minsize()))
            continue;


        // >>> check all rules of the current filter
        const unsigned int rule_count = filter.get_rules().size();
        
        for (unsigned int current_rule = 0;
             current_rule < rule_count;
             current_rule++)
        {
            // get next rule
            const Murx::Rule& rule = filter.get_rules().at(current_rule);
            
            bool rule_matched = false;
            if (rule.negativity)
                rule_matched = true;
            
            // >>> header line filter
            if (! rule.only_check_body_lines)
            {

                // >>> check all header lines whether the rule matches or not
                for (unsigned int current_hl = 0;
                     current_hl < header.get_header_lines().size();
                     current_hl++)
                {
                    // now we try to match the rule to the current header line
                    if (regexec
                        (&rule.compiled_expression,
                         header.get_header_lines().at(current_hl).c_str(),
                         1, &pmatch, 0) == 0)
                    {
                        // MATCH!
                        // -> don't try to match more header lines
                        // -> set 'rule_matched' to 'true' and break the
                        //    'header lines loop'
                        if (! rule.negativity)
                        {
                            // save the match data
                            tmp_match_list.add(Murx::MatchData
                                               (current_hl,
                                                pmatch.rm_so,
                                                pmatch.rm_eo - pmatch.rm_so,
                                                rule.expression,
                                                false,
                                                true));
                            
                            rule_matched = true;
                        }
                        else
                        {
                            tmp_match_list.add(Murx::MatchData(rule.expression));

                            rule_matched = false;
                        }
                        
                        break;
                    }
                }
                // <<< check all header lines whether the rule matches or not
                
            // <<< header line filter
            }
            else
            {
            // >>> body line filter
                
                // check preferences whether to check bodylines of the mail or not
                // -> if so, do the same like with the headerlines
                if (preferences.get_body_lines() > 0
                    && header.get_body_lines().size() > 0)
                {
                    // >>> check all body lines whether the rule matches or not
                    for (unsigned int current_bl = 0;
                         current_bl < header.get_body_lines().size();
                         current_bl++)
                    {
                        // now we try to match the rule to the current header line
                        if (regexec
                            (&rule.compiled_expression,
                             header.get_body_lines().at(current_bl).c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more body lines
                            // -> set 'rule_matched' to 'true' and break the
                            // 'body lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_bl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    false,
                                                    false));
                                
                                rule_matched = true;
                            }
                            else
                            {
                                tmp_match_list.add(Murx::MatchData(rule.expression));

                                rule_matched = false;
                            }
                            
                            break;
                        }
                    }
                    // <<< check all body lines whether the rule matches or not
                }
            }
            // <<< body line filter
        
            // do we have a match for this rule?
            // if not: the whole filter can't match
            // -> set 'filter_matched' to 'false' and break the 'rules loop'
            if (! rule_matched)
            {
                filter_matched = false;
                break;
            }

        }
        // <<< check all rules of the current filter
        
        // if the filter matched: no more filtering is needed
        if (filter_matched)
        {
            // set the testflag in the header for later usage
            header.set_testmode(filter.get_testmode());
            
            // We set both sizes. It does not matter
            // if one of the values is zero
            header.set_maxsize(filter.get_maxsize());
            header.set_minsize(filter.get_minsize());

            // save matches for message
            tmp_match_list.set_action ("ALLOW");
            tmp_match_list.set_complete_match (true);
            header.add_match(tmp_match_list);
            
            // set action to ALLOW here and check for MAXSIZE_ALLOW below
            header.set_action(Murx::Session::ALLOW);

            // PERFORM THE GLOBAL SIZE CHECKS ONLY IF THERE ARE NO LOCAL ONES
            // check whether the mail is too large to be allowed or not
            // if YES: mark mail as 'delete'
            // if  NO: mark mail as 'allowed'
            if ((preferences.get_maxsize_allow() > 0)
                && (filter.get_maxsize() == 0)
                && (filter.get_minsize() == 0))
            {
                if (header.get_size() > preferences.get_maxsize_allow())
                    header.set_action(Murx::Session::DELETE_MAXSIZE_ALLOW);
            }
            
            return true;
        }

        // check if something matched
        if (tmp_match_list.size() > 0)
        {
            // save matches for message
            tmp_match_list.set_action ("ALLOW");
            tmp_match_list.set_complete_match (false);
            header.add_match(tmp_match_list);
        }
    }
    // <<< check all filters of the account
    
    return false;
}
//
// <<< ALLOWS
//



//
// >>> DENIES
//
bool Murx::Session::check_denies (Murx::Header &header)
{    
    // how many deny filters are available?
    const unsigned int filter_count =
        preferences.get_accounts().at(current_account).get_denies().size();
    
    // >>> check all filters of the account
    for (unsigned int current_filter = 0;
         current_filter < filter_count;
         current_filter++)
    {
        // create new vector for temporary saving of the matches
        Murx::Match tmp_match_list;
        
        // get next filter object
        const Murx::Filter& filter =
            preferences.get_accounts().at(current_account).get_denies().at(current_filter);
        
        bool filter_matched = true;
        
        // >>> check all rules of the current filter
        const unsigned int rule_count = filter.get_rules().size();
        
        for (unsigned int current_rule = 0;
             current_rule < rule_count;
             current_rule++)
        {
            // get next rule
            const Murx::Rule &rule = filter.get_rules().at (current_rule);

            bool rule_matched = false;
            if (rule.negativity)
                rule_matched = true;
            
            // >>> header line filter
            if (! rule.only_check_body_lines)
            {

                // >>> check all header lines whether the rule matches or not
                for (unsigned int current_hl = 0;
                     current_hl < header.get_header_lines().size();
                     current_hl++)
                {
                    // now we try to match the rule to the current header line
                    if (regexec
                        (&rule.compiled_expression,
                         header.get_header_lines().at(current_hl).c_str(),
                         1, &pmatch, 0) == 0)
                    {
                        // MATCH!
                        // -> don't try to match more header lines
                        // -> set 'rule_matched' to 'true' and break the
                        //    'header lines loop'
                        if (! rule.negativity)
                        {
                            // save the match data
                            tmp_match_list.add(Murx::MatchData
                                               (current_hl,
                                                pmatch.rm_so,
                                                pmatch.rm_eo - pmatch.rm_so,
                                                rule.expression,
                                                false,
                                                true));
                            
                            rule_matched = true;
                        }
                        else
                            rule_matched = false;
                        
                        break;
                    }
                    
                    
                    // the expression did not match
                    // -> check if the current header line is the subject line
                    // -> if so, normalize the subject and try to match again
                    if (    preferences.get_normalize_subject()
                            &&    (header.get_subject_line_index() > -1)
                            &&    (current_hl == static_cast<unsigned int>
                                   (header.get_subject_line_index()))    )
                    {
                        std::string normalized_subject
                            = header.get_header_lines().at(current_hl);
                        
                        Murx::Utils::normalize_subject(normalized_subject);
                        
                        if (regexec
                            (&rule.compiled_expression,
                             normalized_subject.c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more header lines
                            // -> set 'rule_matched' to 'true' and break the
                            //    'header lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_hl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    true,
                                                    true));
                                
                                rule_matched = true;
                            }
                            else
                                rule_matched = false;
                            
                            break;
                        }
                    }
                }
                // <<< check all header lines whether the rule matches or not

            // <<< header line filter
            }
            else
            {
            // >>> body line filter
 
            // check preferences whether to check bodylines of the mail or not
            // -> if so, do the same like with the headerlines
                if (preferences.get_body_lines() > 0
                    && header.get_body_lines().size() > 0)
                {
                    // >>> check all body lines whether the rule matches or not
                    for (unsigned int current_bl = 0;
                         current_bl < header.get_body_lines().size();
                         current_bl++)
                    {
                        // now we try to match the rule to the current header line
                        if (regexec
                            (&rule.compiled_expression,
                             header.get_body_lines().at(current_bl).c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more body lines
                            // -> set 'rule_matched' to 'true' and break the
                            // 'body lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_bl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    false,
                                                    false));
                                
                                rule_matched = true;
                            }
                            else
                                rule_matched = false;
                            
                            break;
                        }
                    }
                    // <<< check all body lines whether the rule matches or not
                }
            }
            // <<< body line filter
 
            // do we have a match for this rule?
            // if not: the whole filter can't match
            // -> set 'filter_matched' to 'false' and break the 'rules loop'
            if (! rule_matched)
            {
                filter_matched = false;
                break;
            }

            // if the rule matched
            // -> check whether it was a negativity rule and if so,
            //    save the expression
            if (rule.negativity)
            {
                tmp_match_list.add(Murx::MatchData(rule.expression));
            }
        }
        // <<< check all rules of the current filter
        
        if (filter_matched)
        {
            // set the testflag in the header for later usage
            header.set_testmode(filter.get_testmode());
            
            // check local maxsize and minsize
            if ((filter.get_maxsize() != 0
                 &&  header.get_size() >= filter.get_maxsize())
                || (filter.get_minsize() != 0
                    && header.get_size() <= filter.get_minsize()))
            {
                only_size_not_match = true;
                filter_matched = false;
            }
            else
            {
                // We set both sizes. It does not matter
                // if one of the values is zero
                header.set_maxsize(filter.get_maxsize());
                header.set_minsize(filter.get_minsize());

                // save matches for message
                tmp_match_list.set_action ("DENY");
                tmp_match_list.set_complete_match (true);
                header.add_match(tmp_match_list);
            
                // if the filter matched: no more filtering is needed
                header.set_action(Murx::Session::DELETE_DENYRULE);

                // check for testmode
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // decrease message count of this account
                    remaining_message_count--;
                }

                return true;
            }
        }

        // check if something matched
        if (tmp_match_list.size() > 0)
        {
            // save matches for message
            tmp_match_list.set_action ("DENY");
            tmp_match_list.set_complete_match (false);
            header.add_match(tmp_match_list);
        }
    }
    // <<< check all filters of the account

    return false;
}
//
// <<< DENIES
//



//
// >>> MOVETOS
//
bool Murx::Session::check_movetos (Murx::Header &header)
{    
    // how many moveto filters are available?
    const unsigned int filter_count =
        preferences.get_accounts().at(current_account).get_movetos().size();
    
    // >>> check all filters of the account
    for (unsigned int current_filter = 0;
         current_filter < filter_count;
         current_filter++)
    {
        // create new vector for temporary saving of the matches
        Murx::Match tmp_match_list;
        
        // get next filter object
        const Murx::Filter &filter =
            preferences.get_accounts().at (current_account).get_movetos().at (current_filter);
        
        bool filter_matched = true;
        
        // >>> check all rules of the current filter
        const unsigned int rule_count = filter.get_rules().size();
        
        for (unsigned int current_rule = 0;
             current_rule < rule_count;
             current_rule++)
        {
            // get next rule
            const Murx::Rule &rule = filter.get_rules().at (current_rule);

            bool rule_matched = false;
            if (rule.negativity)
                rule_matched = true;
            
            // >>> header line filter
            if (! rule.only_check_body_lines)
            {

                // >>> check all header lines whether the rule matches or not
                for (unsigned int current_hl = 0;
                     current_hl < header.get_header_lines().size();
                     current_hl++)
                {
                    // now we try to match the rule to the current header line
                    if (regexec
                        (&rule.compiled_expression,
                         header.get_header_lines().at(current_hl).c_str(),
                         1, &pmatch, 0) == 0)
                    {
                        // MATCH!
                        // -> don't try to match more header lines
                        // -> set 'rule_matched' to 'true' and break the
                        //    'header lines loop'
                        if (! rule.negativity)
                        {
                            // save the match data
                            tmp_match_list.add(Murx::MatchData
                                               (current_hl,
                                                pmatch.rm_so,
                                                pmatch.rm_eo - pmatch.rm_so,
                                                rule.expression,
                                                false,
                                                true));

                            rule_matched = true;
                        }
                        else
                            rule_matched = false;
                    
                        break;
                    }
                    
                
                    // the expression did not match
                    // -> check if the current header line is the subject line
                    // -> if so, normalize the subject and try to match again
                    if (preferences.get_normalize_subject()
                        && (header.get_subject_line_index() > -1)
                        && (current_hl == static_cast<unsigned int>
                            (header.get_subject_line_index())))
                    {
                        std::string normalized_subject
                            = header.get_header_lines().at(current_hl);

                        Murx::Utils::normalize_subject(normalized_subject);
                    
                        if (regexec
                            (&rule.compiled_expression,
                             normalized_subject.c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more header lines
                            // -> set 'rule_matched' to 'true' and break the
                            //    'header lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_hl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    true,
                                                    true));

                                rule_matched = true;
                            }
                            else
                                rule_matched = false;
                        
                            break;
                        }
                    }
                }
                // <<< check all header lines whether the rule matches or not

            // <<< header line filter
            }
            else
            {
            // >>> body line filter 
            
                // check preferences whether to check bodylines of the mail or not
                // -> if so, do the same like with the headerlines
                if (preferences.get_body_lines() > 0
                    && header.get_body_lines().size() > 0)
                {
                    // >>> check all body lines whether the rule matches or not
                    for (unsigned int current_bl = 0;
                         current_bl < header.get_body_lines().size();
                         current_bl++)
                    {
                        // now we try to match the rule to the current header line
                        if (regexec
                            (&rule.compiled_expression,
                             header.get_body_lines().at(current_bl).c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more body lines
                            // -> set 'rule_matched' to 'true' and break the
                            // 'body lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_bl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    false,
                                                    false));

                                rule_matched = true;
                            }
                            else
                                rule_matched = false;
                        
                            break;
                        }
                    }
                    // <<< check all body lines whether the rule matches or not
                }
            }
            //  <<< body line filter
            
            // do we have a match for this rule?
            // if not: the whole filter can't match
            // -> set 'filter_matched' to 'false' and break the 'rules loop'
            if (! rule_matched)
            {
                filter_matched = false;
                break;
            }

            // if the rule matched
            // -> check whether it was a negativity rule and if so,
            //    save the expression
            if (rule.negativity)
            {
                tmp_match_list.add(Murx::MatchData(rule.expression));
            }
        }
        // <<< check all rules of the current filter
        
        if (filter_matched)
        {
            // set the testflag in the header for later usage
            header.set_testmode(filter.get_testmode());
            
            // check local maxsize and minsize
            if ((filter.get_maxsize() != 0
                 && header.get_size() >= filter.get_maxsize())
                || (filter.get_minsize() != 0
                    && header.get_size() <= filter.get_minsize()))
            {
                only_size_not_match = true;
                filter_matched = false;
            }
            else
            {
                // We set both sizes. It does not matter
                // if one of the values is zero
                header.set_maxsize(filter.get_maxsize());
                header.set_minsize(filter.get_minsize());

                // set the appropriate action
                std::string action = "MOVETO \""
                    + static_cast<const Murx::MoveTo&> (filter).get_folder()
                    + "\"";

                tmp_match_list.set_action (action);
                tmp_match_list.set_complete_match (true);

                // save matches for message
                header.add_match(tmp_match_list);
            
                // if the filter matched: no more filtering is needed
                header.set_action(Murx::Session::MOVETO);
                header.set_folder
                    (static_cast<const Murx::MoveTo&> (filter).get_folder());

                /*
                // check for testmode
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // decrease message count of this account
                    remaining_message_count--;
                }
                */

                return true;
            }
        }

        // check if something matched
        if (tmp_match_list.size() > 0)
        {
            // set the appropriate action
            std::string action = "MOVETO \""
                + static_cast<const Murx::MoveTo&> (filter).get_folder()
                + "\"";
            
            tmp_match_list.set_action (action);
            tmp_match_list.set_complete_match (false);
            
            // save matches for message
            header.add_match(tmp_match_list);
        }
    }
    // <<< check all filters of the account

    return false;
}
//
// <<< MOVETOS
//



//
// >>> MAXSIZE_DENY
//
bool Murx::Session::check_maxsize (Murx::Header &header)
{
    Murx::Preferences& preferences = *Murx::Preferences::get_instance();

    // check if the mail is too large (MAXSIZE_DENY)
    if (! only_size_not_match
        &&  (preferences.get_maxsize_deny() > 0)
        &&  (header.get_size() > preferences.get_maxsize_deny()))
    {
        header.set_action(Murx::Session::DELETE_MAXSIZE_DENY);
        
        // check for testmode
        if (!preferences.get_testmode() && !header.get_testmode())
        {
            // decrease message count of this account
            remaining_message_count--;
        }
        
        return true;
    }

    return false;
}
//
// <<< MAXSIZE_DENY
//



//
// >>> SCORES
//
bool Murx::Session::check_scores (Murx::Header &header)
{
    // the current score of this message
    int score = 0;
    
    // how many score filters are available?
    const unsigned int filter_count =
        preferences.get_accounts().at(current_account).get_scores().size();
    
    // >>> check all filters of the account
    for (unsigned int current_filter = 0;
         current_filter < filter_count;
         current_filter++)
    {
        // create new vector for temporary saving of the matches
        Murx::Match tmp_match_list;
        
        // get next filter object
        const Murx::Filter &filter =
            preferences.get_accounts().at (current_account).get_scores().at (current_filter);
        
        bool filter_matched = true;


        // check the local sizes

        // maximum size
        if (filter.get_maxsize() != 0
            && (header.get_size() >= filter.get_maxsize()))
            continue;

        // minimal size
        if (filter.get_minsize() != 0
            && (header.get_size() <= filter.get_minsize()))
            continue;

        
        // >>> check all rules of the current filter
        const unsigned int rule_count = filter.get_rules().size();
        
        for (unsigned int current_rule = 0;
             current_rule < rule_count;
             current_rule++)
        {
            // get next rule
            const Murx::Rule &rule = filter.get_rules().at (current_rule);
            
            bool rule_matched = false;
            if (rule.negativity)
                rule_matched = true;

            // >>> header line filter
            if (! rule.only_check_body_lines)
            {
            
                // >>> check all header lines whether the rule matches or not
                for (unsigned int current_hl = 0;
                     current_hl < header.get_header_lines().size();
                     current_hl++)
                {
                    // now we try to match the rule to the current header line
                    if (regexec
                        (&rule.compiled_expression,
                         header.get_header_lines().at(current_hl).c_str(),
                         1, &pmatch, 0) == 0)
                    {
                        // MATCH!
                        // -> don't try to match more header lines
                        // -> set 'rule_matched' to 'true' and break the
                        //    'header lines loop'
                        if (! rule.negativity)
                        {
                            // save the match data
                            tmp_match_list.add(Murx::MatchData
                                               (current_hl,
                                                pmatch.rm_so,
                                                pmatch.rm_eo - pmatch.rm_so,
                                                rule.expression,
                                                false,
                                                true));
                            rule_matched = true;
                        }
                        else
                            rule_matched = false;
                    
                        break;
                    }
                
                    // the expression did not match
                    // -> check if the current header line is the subject line
                    // -> if so, normalize the subject and try to match again
                    if (preferences.get_normalize_subject()
                        && (header.get_subject_line_index() > -1)
                        && (current_hl == static_cast<unsigned int>
                            (header.get_subject_line_index())))
                    {
                        std::string normalized_subject =
                            header.get_header_lines().at(current_hl);

                        Murx::Utils::normalize_subject(normalized_subject);
                    
                        if (regexec
                            (&rule.compiled_expression,
                             normalized_subject.c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more header lines
                            // -> set 'rule_matched' to 'true' and break the
                            //    'header lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_hl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    true,
                                                    true));

                                rule_matched = true;
                            }
                            else
                                rule_matched = false;
                        
                            break;
                        }
                    }
                }
                // <<< check all header lines whether the rule matches or not

            // <<< header line filter
            }
            else
            {
            // >>> body line filter

                // check preferences whether to check bodylines of the mail or not
                // -> if so, do the same like with the headerlines
                if (preferences.get_body_lines() > 0
                    && header.get_body_lines().size() > 0)
                {
                    // >>> check all body lines whether the rule matches or not
                    for (unsigned int current_bl = 0;
                         current_bl < header.get_body_lines().size();
                         current_bl++)
                    {
                        // now we try to match the rule to the current header line
                        if (regexec
                            (&rule.compiled_expression,
                             header.get_body_lines().at(current_bl).c_str(),
                             1, &pmatch, 0) == 0)
                        {
                            // MATCH!
                            // -> don't try to match more body lines
                            // -> set 'rule_matched' to 'true' and break the
                            // 'body lines loop'
                            if (! rule.negativity)
                            {
                                // save the match data
                                tmp_match_list.add(Murx::MatchData
                                                   (current_bl,
                                                    pmatch.rm_so,
                                                    pmatch.rm_eo - pmatch.rm_so,
                                                    rule.expression,
                                                    false,
                                                    false));

                                rule_matched = true;
                            }
                            else
                                rule_matched = false;

                            break;
                        }
                    }
                    // <<< check all body lines whether the rule matches or not
                }
            }
            // <<< body line filter
   
            // do we have a match for this rule?
            // if not: the whole filter can't match
            // -> set 'filter_matched' to 'false' and break the 'rules loop'
            if (! rule_matched)
            {
                filter_matched = false;
                break;
            }
            // if the rule matched
            // -> check whether it was a negativity rule and if so,
            //    save the expression
            if (rule.negativity)
            {
                tmp_match_list.add(Murx::MatchData(rule.expression));
            }
        }
        // <<< check all rules of the current filter
        
        // if the filter matched:
        //  add the score-value if no test bit is set
        if (filter_matched)
        {
            //
            // ****   For further description read NOTE below.   ****
            //
            // only check for local testmode
            if (! filter.get_testmode())
            {
                // nice, the current filter matched
                // now add the filter score to the local score
                score += static_cast<const Murx::Score&> (filter).get_score();
            }

            // We set both sizes. It does not matter
            // if one of the values is zero
            header.set_maxsize(filter.get_maxsize());
            header.set_minsize(filter.get_minsize());

            // set the appropriate action
            std::string action = "SCORE "
                + Murx::Utils::int_to_string
                (static_cast<const Murx::Score&> (filter).get_score());

            tmp_match_list.set_action (action);
            tmp_match_list.set_complete_match (true);

            // save matches for message
            header.add_match(tmp_match_list);
        }
        else
        {
            // check if something matched
            if (tmp_match_list.size() > 0)
            {
                // set the appropriate action
                std::string action = "SCORE "
                    + Murx::Utils::int_to_string
                    (static_cast<const Murx::Score&> (filter).get_score());
                
                tmp_match_list.set_action (action);
                tmp_match_list.set_complete_match (false);
                
                // save matches for message
                header.add_match(tmp_match_list);
            }
        }
    }
    // <<< check all filters of the account
    
    
    // used for logging in murx.cxx
    header.set_total_score(score);


    // now we have to check whether the score of the header is
    // 1. greater than HIGHSCORE (DELETE)
    // 2. less than HIGHSCORE (DO NOTHING)
    if (score >= preferences.get_highscore())
    {
        header.set_action(Murx::Session::SCORE_DELETE);

        // NOTE:
        //
        // Testmode for SCORE rules is slightly different from the testmode
        // in ALLOW, DENY and MOVETO rules.
        //
        // If the global testmode is set, a message reaching the highscore
        // will not get deleted. This is because the ACTION for HIGHSCORE is
        // DELETE.
        //
        // The local testmode indicates that a SCORE should not be added to
        // the message's overall score. The ACTION is ADD_SCORE not DELETE.
        //
        // only check for global testmode
        if (! preferences.get_testmode())
        {
            // decrease message count of this account
            remaining_message_count--;
        
            //delete_mail (header.get_account (), header.get_id ());
        }

        return true;
    }
    else
        return false;
}
//
// <<< SCORES
//


void Murx::Session::delete_mail (unsigned int account, unsigned int id)
{
    try
    {
        protocol.at(account)->delete_message(id);
    }
    catch (std::exception &e)
    {
        // TODO: STL exception possible because of bad vector index
        std::cerr << "Murx::Session::delete_mail(): STL exception caught: "
                  << e.what ()
                  << std::endl;
    }
}


void Murx::Session::move_mail (unsigned int account, unsigned int id,
                               const std::string& folder)
{
    try
    {
        if (Murx::IMAPProtocol::IMAPProtocol* p
            = dynamic_cast<IMAPProtocol*> (protocol.at(account)))
        {
            // p is of type IMAPProtocol
            p->move_message(id, folder);
        }
        else
        {
            std::string msg =
                _("MOVETO command not available in non-IMAP account \"")
                + preferences.get_accounts().at(current_account).get_name()
                + "\"";

            throw Murx::Exceptions::CommandNotAvailable(msg);
        }
    }
    catch (std::exception &e)
    {
        // TODO: STL exception possible because of bad vector index
        std::cerr << "Murx::Session::move_mail(): STL exception caught: "
                  << e.what ()
                  << std::endl;
    }
}


int Murx::Session::get_current_account_num ()
{
    return current_account + 1;
}
int Murx::Session::get_all_accounts_num () const
{
    return preferences.get_accounts().size();
}


void Murx::Session::finish ()
{
    for (unsigned int i = 0; i < protocol.size(); i++)
    {
        protocol[i]->logout();
        delete protocol[i];
    }

    while (protocol.size() > 0)
        protocol.pop_back();
}


void Murx::Session::finish_current_account ()
{
    if (protocol.size() > 0)
    {
        protocol[protocol.size() - 1]->logout();
        delete protocol[protocol.size() - 1];
        protocol.pop_back();
    }
}
