// Session.hxx - part of the murx project
//
// (c) 2004 by Thomas Stauss, Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifndef MURX_SESSION_HXX
#define MURX_SESSION_HXX

#include <string>
#include <vector>
#include <fstream>
extern "C"
{
#if USE_PCRE
#  include <pcreposix.h>   // regmatch_t
#else
#  include <regex.h>       // regmatch_t
#endif
}

#include "Header.hxx"
#include "Protocol.hxx"
#include "Account.hxx"
#include "Preferences.hxx"
#include "RCParser.hxx"
#include "RFC822Parser.hxx"
#include "Filter.hxx"
#include "MoveTo.hxx"
#include "Score.hxx"


namespace Murx
{
    class Session
    {
    private:
        Preferences&              preferences;

        std::vector<Protocol*>    protocol;
        //std::vector<Account*>*    accounts;

        std::vector<std::string>  message_ids;

        regmatch_t                pmatch;
        bool                      want_next_account;
        //unsigned int              account_count;
        unsigned int              current_account;
        
        //unsigned int              filter_count;
        //unsigned int              rule_count;

        unsigned int              remaining_message_count;

        bool                      first_run;

        bool                      only_size_not_match;
        
        bool check_conformance (Header &header);
        bool check_maxlength   (Header &header);
        bool check_duplicates  (Header &header);
        bool check_allows      (Header &header);
        bool check_denies      (Header &header);
        bool check_movetos     (Header &header);
        bool check_maxsize     (Header &header);
        bool check_scores      (Header &header);
        
    public:
        // >>> STATUS FLAGS
        // use these flags to determine the result of the filter process
        // the flag is placed in the header-object
        static const int ACCOUNT_FILTERING_FINISHED;
        static const int NORULE;
        static const int ALLOW;
        static const int DELETE_MAXSIZE_ALLOW;
        static const int DELETE_DENYRULE;
        static const int DELETE_MAXSIZE_DENY;
        static const int DELETE_DUPLICATE;
        static const int DELETE_NONCONFORMANT;
        static const int DELETE_MAXLENGTH_EXCEEDED;
        static const int MOVETO;
        static const int MOVETO_NONCONFORMANT;
        static const int SCORE_DELETE;
        // <<< STATUS FLAGS
        
        // initializes a new session object
        // 'rcfile' = filename of rcfile or "$HOME/.murxrc" if 'rcfile' is empty
        Session (std::string rcfile);
        ~Session();
        
        // call this method to filter the next mail
        //
        // * EACH CALL HAVE TO BE DONE WITH A CLEAN, NEWLY CREATED HEADER-OBJECT
        // * YOU HAVE TO PASS A POINTER TO A 'REAL' HEADER-OBJECT,
        // * CREATED WITH 'new Header ...'
        //
        // return values:    TRUE: a mail was available, 'header' is valid
        //                  FALSE: no more mails available, 'header' is invalid
        //
        // if the statusflag is set to 'ACCOUNT_FILTERING_FINISHED',
        // call 'filter_next_mail' again to check next mail account
        //
        bool           filter_next_mail            (Header& header);
        void           account_back                ();
        
        const Account& get_current_account         () const;
        
        std::string    get_current_accountname     () const;
        std::string    get_next_accountname        ();

        int            get_current_message_count   () const;
        
        int            get_current_account_num     ();
        int            get_all_accounts_num        () const;

        unsigned int   get_remaining_message_count () const;
        
        void           delete_mail                 (unsigned int account,
                                                    unsigned int id);
        void           delete_all_mails            ();
        
        void           move_mail                   (unsigned int account,
                                                    unsigned int id,
                                                    const std::string& folder);

        void           finish                      ();
        void           finish_current_account      ();
    };
}

#endif /* MURX_SESSION_HXX */
