// TCPSocket.cxx - part of the murx project
//
// (c) 2004 by Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifdef WIN32
    #include <Winsock2.h>
#else
    #include <sys/types.h>
    #include <sys/socket.h>
    #include <netdb.h>
    #include <netinet/in.h>
    #include <arpa/inet.h>
    #include <unistd.h>
#endif

#include <string>
//#include <iostream>
#include <csetjmp>

#include "Socket.hxx"
#include "TCPSocket.hxx"
#include "Exceptions.hxx"
#include "i18n.hxx"

#include "SignalHandler.hxx"
#include "SIGALRMHandler.hxx"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


Murx::Net::TCPSocket::TCPSocket (std::string server,
                                 unsigned int port,
                                 unsigned int timeout)
    : connected(false)
{
    // store given timeout
    this->timeout = timeout;

#ifdef WIN32
    WSADATA wsaData;
    WSAStartup(2, &wsaData);
#endif

    serverInfo.sin_family = AF_INET;

#ifdef WIN32
    if (inet_aton (server.c_str (), &(serverInfo.sin_addr)) == 0)
    {
#endif

#ifndef WIN32
        //
        // timeout for gethostbyname () and connect ()
        //
        //   NOTE:
        //
        //   This method relies on calling [Murx::Socket::]connect () directly
        //   after creating a new Murx::Socket. The timer will run after having
        //   created a new Socket.
        //


        // install a handler for timeouts
        Murx::SIGALRMHandler sigalrm_handler;

        Murx::SignalHandler::get_instance()->register_handler
            (SIGALRM, &sigalrm_handler);


        // check for timeout
        if (sigsetjmp (*sigalrm_handler.get_curr_env(), 1) == 0)
        {
            // set timeout
            alarm (timeout);
#endif
            struct hostent *host;
            
            host = gethostbyname (server.c_str ());
            
            if (host == 0)
            {
#ifndef WIN32
                // reset timeout
                alarm (0);
                Murx::SignalHandler::get_instance()->remove_handler(SIGALRM);
#endif
                throw Murx::Exceptions::UnknownHostname ();
            }
            
            serverInfo.sin_addr = *(struct in_addr*) host->h_addr;
            
#ifdef WIN32
        }
#endif
        
        serverInfo.sin_port = htons (port);
        
        sock = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
        
#ifdef WIN32
        if (sock == INVALD_SOCKET)
#else
            if (sock == -1)
#endif
            {
#ifndef WIN32
                // reset timeout
                alarm (0);
                Murx::SignalHandler::get_instance()->remove_handler(SIGALRM);
#endif
                throw Murx::Exceptions::NoSocket ();
            }
#ifndef WIN32
    }  // if (sigsetjmp (*sigalrm_handler.get_curr_env(), 1) == 0)
    else
    {
        // reset timeout
        alarm (0);
        Murx::SignalHandler::get_instance()->remove_handler(SIGALRM);

        // timeout!
        throw Murx::Exceptions::Timeout ();
    }
#endif
}


void Murx::Net::TCPSocket::connect ()
{
    // if we are already connected: do nothing!
    if (connected)
        return;
    
    if (::connect (sock,
                   (struct sockaddr*) &serverInfo,
                   sizeof(serverInfo)) != 0)
        {
#ifndef WIN32
            // reset timeout
            alarm (0);
            Murx::SignalHandler::get_instance()->remove_handler(SIGALRM);
#endif
            throw Murx::Exceptions::NoConnection ();
        }
    
#ifndef WIN32
    //
    // the timeout after connecting to the server is handled via select () 
    //
    // reset timeout
    alarm (0);
    Murx::SignalHandler::get_instance()->remove_handler(SIGALRM);
#endif

    connected = true;
}


void Murx::Net::TCPSocket::close ()
{
    if (connected)
    {
#ifdef WIN32
        closesocket (sock);
#else
        ::close (sock);
#endif
        
        connected = false;
    }
}


// TODO: Install signal-handler to catch SIGPIPE
// If we have no such handler, the process will
// get killed if we send to a closed connection
//

void Murx::Net::TCPSocket::write (const std::string str)
{
    if (send (sock, str.c_str(), str.length(), 0) == -1)
        throw Murx::Exceptions::ConnectionLost ();
}


void Murx::Net::TCPSocket::write (const char *str, int len)
{
    if (send (sock, str, len, 0) == -1)
        throw Murx::Exceptions::ConnectionLost ();
}


void Murx::Net::TCPSocket::write (const char *str)
{
    if (send (sock, str, strlen (str), 0) == -1)
        throw Murx::Exceptions::ConnectionLost ();
}


int Murx::Net::TCPSocket::read (std::string &str)
{
    const int buf_size = 4095;
    char buf[buf_size+1];
    int len;
    int retval;
    
    // zero the descriptor-set
    FD_ZERO (&fds);
    // add the socket to the descriptor-set
    FD_SET (sock, &fds);
    // set the timeout
    tv.tv_sec = timeout;  // sec
    tv.tv_usec = 0; // usec
    
    // wait until data is available or we have a timeout
    retval = select (sock+1, &fds, 0, 0, &tv);
    
    // something has gone wrong with the connection
    if (retval == -1)
    {
        throw Murx::Exceptions::ConnectionLost ();
    }
    // no data is available -> timeout
    if (retval == 0)
    {
        throw Murx::Exceptions::Timeout ();
    }
    
    // no error, no timeout: data is available
    
    // receive data from server, max. buf_size bytes
    len = recv (sock, buf, buf_size, 0);
    // if we could not receive data, there is an error
    if (len < 1)
    {
        // close this socket
        this->close ();
        // inform the upper layers about this error
        throw Murx::Exceptions::ConnectionLost ();
    }
    
    // terminate the fetched data
    buf[len] = '\0';
    // save fetched data in 'str'
    str += buf;
    
    // return the count of bytes read
    return len;
}


int Murx::Net::TCPSocket::get_socket ()
{
    return sock;
}


bool Murx::Net::TCPSocket::is_connected ()
{
    return connected;
}


unsigned int Murx::Net::TCPSocket::get_timeout ()
{
    return timeout;
}
