// TLSSocket.cxx - part of the murx project
//
// (c) 2004 by Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if USE_OPENSSL

#ifdef HAVE_OPENSSL_SSL_H
#include <openssl/ssl.h>
#endif

#include <iostream>
#include <string>
#include <sstream>

#include "TCPSocket.hxx"
#include "Exceptions.hxx"
#include "TLSSocket.hxx"
#include "i18n.hxx"


Murx::Net::TLSSocket::TLSSocket (const std::string server,
                                 const unsigned int port,
                                 unsigned int timeout,
                                 std::string fingerprint)
{
    sock = new Murx::Net::TCPSocket (server, port, timeout);
    this->fingerprint = fingerprint;

    tls_init ();
}


Murx::Net::TLSSocket::TLSSocket (Murx::Net::TCPSocket *sock,
                                 std::string fingerprint)
{
    this->sock = sock;
    this->fingerprint = fingerprint;

    tls_init ();
}


Murx::Net::TLSSocket::~TLSSocket ()
{
    delete sock;
}


void Murx::Net::TLSSocket::tls_init ()
{
    // Initialize the OpenSSL library
    SSL_library_init ();

    meth = TLSv1_client_method ();

    if ((ssl_ctx = SSL_CTX_new (meth)) == 0)
        throw Murx::Exceptions::NoSSLContext ();
    
    if ((ssl = SSL_new (ssl_ctx)) == 0)
        throw Murx::Exceptions::SSL (_("Could not create SSL structure"));
}


void Murx::Net::TLSSocket::connect ()
{
    if (!sock->is_connected ())
        sock->connect ();

    if (SSL_set_fd (ssl, sock->get_socket ()) != 1)
    {
        throw Murx::Exceptions::SSL (
            _("Could not connect SSL object to socket"));
    }

    if (SSL_connect (ssl) != 1)
        throw Murx::Exceptions::SSLHandshake ();
    
    
    // calculate fingerprints of server-certificate
    std::string fpr_md5 = get_fingerprint ("MD5");
    std::string fpr_sha1 = get_fingerprint ("SHA1");
    std::string *cur_fpr;
    
    
    // There is no fingerprint given:
    // Provide one to the user, but close connection for now
    if (fingerprint.empty ())
    {
        // close connection
        this->close ();
        
        // provide fingerprints to user
        throw Murx::Exceptions::NoFingerprint (fpr_md5, fpr_sha1);
    }
    
    // detect which kind of fingerprint we have to check
    if (fingerprint.length () == 47)
        cur_fpr = &fpr_md5;
    else
        cur_fpr = &fpr_sha1;
    
    // check if the certificate is unchanged
    if (fingerprint != *cur_fpr)
    {    
        // Oops, the certificate seems to have changed
        
        // shutdown connection
        this->close ();
        
        // provide new fingerprints to the user
        throw Murx::Exceptions::FingerprintChanged (fpr_md5, fpr_sha1);
    }
}


void Murx::Net::TLSSocket::close ()
{
    SSL_shutdown (ssl);
    // 'SSL_shutdown' has to be called twice!
    // otherwise a packet with 'RST'-flag will be sent
    SSL_shutdown (ssl);
    sock->close ();
}


void Murx::Net::TLSSocket::write (std::string str)
{
    SSL_write (ssl, str.c_str(), str.length());
}


void Murx::Net::TLSSocket::write (const char *str, int len)
{
    SSL_write (ssl, str, len);
}


void Murx::Net::TLSSocket::write (const char *str)
{
    SSL_write (ssl, str, strlen (str));
}


int Murx::Net::TLSSocket::read (std::string &str)
{
    const int buf_size = 4095;
    char buf[buf_size+1];
    int len;
    int len_complete = 0;
    
    // check if we have something in the SSL-buffer
    // if not, we have to wait...
    if (SSL_pending (ssl) < 1)
    {
        int retval;
        
        // zero the descriptor-set
        FD_ZERO (&fds);
        // add the socket to the descriptor-set
        FD_SET (this->sock->get_socket (), &fds);
        // set the timeout
        tv.tv_sec = this->sock->get_timeout ();  // sec
        tv.tv_usec = 0; // usec
        
        // wait until data is available or we have a timeout
        retval = select(this->sock->get_socket ()+1, &fds, 0, 0, &tv);
        
        // something has gone wrong with the connection
        if (retval == -1)
        {
            throw Murx::Exceptions::ConnectionLost ();
        }
        // no data is available -> timeout
        if (retval == 0)
        {
            throw Murx::Exceptions::Timeout ();
        }
    }
    
    do
    {
        len_complete += len = SSL_read (ssl, buf, buf_size);
        if (len < 1)
            throw Murx::Exceptions::ConnectionLost ();
        
        buf[len] = '\0';
        str += buf;
    }
    while (SSL_pending (ssl) > 0);
    
    
    return len_complete;
}


std::string Murx::Net::TLSSocket::get_fingerprint (std::string type)
{
    unsigned int len;
    unsigned char fingerprint[1024];
    std::ostringstream fpr_str;
    const EVP_MD *digest;
    
    if (type == "MD5")
        digest = EVP_md5 ();
    else
        digest = EVP_sha1 ();
    
    if (X509_digest (SSL_get_peer_certificate (ssl), digest, fingerprint, &len))
    {
        for (unsigned int i = 0; i < len; i++)
        {
            fpr_str.width (2);
            fpr_str.fill ('0');
            fpr_str << std::uppercase << std::hex
                    << static_cast<int>(fingerprint[i]);
            if (i < len-1)
                fpr_str << ":";
        }
    }
    
    return fpr_str.str ();
}


unsigned int Murx::Net::TLSSocket::get_timeout ()
{
    return sock->get_timeout ();
}

#endif /* USE_OPENSSL */
