// Utils.cxx - some useful functions, pulled out of the global namespace
//
// (c) 2004 by Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------
//       headerfile            needed for (function, type)
//
#include <string>
#include <vector>
#include <sstream>
#include <exception>
#include <stdexcept>        // out_of_range
#include <cctype>           // isxdigit(), tolower()
#include <cstdio>           // popen(), pclose()
#include <cstdlib>          // getenv()
#include "i18n.hxx"
#include "Exceptions.hxx"
#include "Utils.hxx"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

extern "C"
{
#include <sys/types.h>      // pid_t
#include <unistd.h>         // getpid ()
}


//
// private
//

std::string Murx::Utils::exec_shell (std::string command, unsigned int line)
{
    FILE        *file;
    const int   BUFFER_SIZE = 256;
    char        buffer[BUFFER_SIZE];
    std::string str;
    
    // initialize buffer
    memset (buffer,0,sizeof (buffer));
    
#ifndef WIN32
    // redirect stderr
    command += " 2>/dev/null";
#else
    // don't know how this works with Windows :-|
#endif
    
    // remove backslashes
    while (1)
    {
        try
        {
            unsigned int pos = command.find ('\\');
            command.replace (pos,1,"");
        }
        catch (std::exception &e)
        {
            // no (more) backslash(es) found
            break;
        }
    }
    
    if ((file = popen (command.c_str (), "r")) != 0)
    {
        fgets (buffer, sizeof (buffer), file);
        
        str = buffer;
        
        if (str.size () > 0)
        {
            if (str[str.size () - 1] == '\n')
            {
                // remove trailing newline
                str = str.substr (0, str.length ()-1);
            }
        }
        else
        {
            std::string message = "(exec_shell) ";
            message += _("error on command execution");
            
            if (line > 0)
                message += _(" in line ") + int_to_string (line+1);
            
            throw (Murx::Exceptions::IOError (message));
        }
    }
    else
    {
        std::string message = "(exec_shell) ";
        message += _("input/output error");
        
        throw (Murx::Exceptions::IOError (message));
    }
    
    if (pclose (file) == -1)
    {
        std::string message = "(exec_shell) ";
        message += _("input/output error");
        
        throw (Murx::Exceptions::IOError (message));
    }
    
    return str;
}


//
// public
//

bool Murx::Utils::check_fingerprint (const std::string fingerprint)
{
    // not sure if the next line is portable
    if (fingerprint.length () != 47 && fingerprint.length () != 59)
    {
        return false;
    }
    
    unsigned int index = 0;
    
    // check for xx
    if (! isxdigit (fingerprint[index++]))
        return false;
    
    if (! isxdigit (fingerprint[index++]))
        return false;
    
    
    // check for :xx, 15 times
    for (int i = 0; i < 15; i++)
    {
        if (fingerprint[index++] != ':')
            return false;
        
        if (! isxdigit (fingerprint[index++]))
            return false;
        
        if (! isxdigit (fingerprint[index++]))
            return false;
    }
            
    if (fingerprint.length () == 47)
        return true;
    
    // check for :xx, 4 times
    for (int i = 0; i < 4; i++)
    {
        if (fingerprint[index++] != ':')
            return false;
        
        if (! isxdigit (fingerprint[index++]))
            return false;
        
        if (! isxdigit (fingerprint[index++]))
            return false;
    }
    
    return true;
}


void Murx::Utils::chop (std::string& str)
{
    if (str.length() > 2)
    {
        // remove leading character
        str.erase (0, 1);
        // remove trailing character
        str.erase (str.length()-1, 1);
    }
    else
    {
        // clear string
        str.erase();
    }
}


void Murx::Utils::expand_cmd_subst (std::string& str, unsigned int line)
{
    int           pos;          // actual position in str
    int           cmd_begin;    // points to the begin of command
    int           cmd_end;      // points to the end of command
    std::string   cmd_def;      // `command`
    std::string   command;      // command
    std::string   cmd_output;   // output of command (first line)
    
    while ((pos = str.find ('`')) != -1)
    {
        cmd_begin = pos + 1;
        cmd_end   = str.find ('`', cmd_begin);
        
        if (cmd_end == -1)
        {
            std::string message = _("delimiter missing");
            
            if (line > 0)
                message += _(" in line ") + int_to_string (line+1);
            
            throw (Murx::Exceptions::DelimiterMissing (message));
        }
        
        command = str.substr (cmd_begin, cmd_end - cmd_begin);
        cmd_def = str.substr (cmd_begin - 1, cmd_end - cmd_begin + 2); // `` = 2 chars
        
        // execute command, store output in cmd_output
        cmd_output = exec_shell (command.c_str (), line);
        
        // replace command substitution in string with output of command
        str.replace (str.find (cmd_def), cmd_def.size (), cmd_output);
    }
}


void Murx::Utils::expand_env_vars (std::string& str, unsigned int line)
{
    int           pos;                 // actual position in str
    unsigned int  varname_begin = 0;   // points to the begin of variable name
    int           varname_end   = -1;  // points to the end of vaiable name
    std::string   vardef;              // ${VAR},$VAR
    std::string   varname;             // VAR
    std::string   varval;              // value of the variable
    bool          with_braces;         // flag to distinguish ${VAR} and $VAR
    pid_t         my_pid;              // my process id
            
    while ((pos = str.find ("$")) != -1)
    {
        if (pos != -1)
        {
            if (pos > 0)
            {
                // check for \$
                if (str[pos-1] == '\\')
                {
                    // leave out
                    continue;
                }
            }
            
            try
            {
                if (str.at (pos+1) == '{' )
                {
                    with_braces = true;
                }
                else
                {
                    with_braces = false;
                }
                
                // variant 1: ${VAR}
                //
                // str[pos]   = $
                // str[pos+1] = {
                // str[pos+2] = anything but }
                // str[pos+3] = first possible pos. of }
                //
                // variant 2: $VAR
                //
                // str[pos]   = $
                // str[pos+1] = [[:alnum:]]
                // str[pos+2] = first possible pos. of delimiter (![[:alnum:]])
                //
                if (with_braces)
                {
                    varname_begin = pos+2;
                    varname_end   = str.find ('}',pos+3);
                }
                else
                {
                    varname_begin = pos+1;
                    
                    // special case: $$ (current process id)
                    if (str[varname_begin] == '$')
                    {
                        my_pid = getpid ();
                            
                        varval = Murx::Utils::int_to_string (my_pid);
                                
                        // replace variable name in string with value
                        str = str.replace (str.find("$$"), 2, varval);
                        
                        continue;
                    }
                    
                    for (unsigned int i = varname_begin; i < str.size(); i++)
                    {
                        // does not matter if no match is found
                        if (!isalnum (str[i]))
                        {
                            varname_end = i;
                            break;
                        }
                    }
                }
                
                if (varname_end == -1)
                {
                    std::string message;
                    
                    if (line > 0)
                        message = _("delimiter missing in line ")
                            + Murx::Utils::int_to_string (line+1);
                    else
                        message = _("delimiter missing");
                    
                    throw (Murx::Exceptions::DelimiterMissing (message));
                }
                
                for (int i = varname_begin; i < varname_end; i++)
                {
                    // check for correct variable name
                    
                    if (!isalnum (str[i]))
                    {
                        std::string message;
                        
                        if (line > 0)
                            message = _("illegal variable identifier in line ")
                                + Murx::Utils::int_to_string (line+1);
                        else
                            message = _("illegal variable identifier");
                        
                        throw (Murx::Exceptions::InvalidVariableIdentifier (message));
                    }
                }
                
                if (with_braces)
                {
                    // ${} = 3 characters
                    vardef = str.substr (pos,varname_end - varname_begin + 3);
                }
                else
                {
                    // $   = 1 character
                    vardef = str.substr (pos,varname_end - varname_begin + 1);
                }
                
                varname =
                    str.substr (varname_begin, varname_end - varname_begin);
                
                if (getenv (varname.c_str ()) != 0)
                {
                    // get value of environment variable
                    varval = getenv (varname.c_str ());
                    
                    // replace variable name in string with value
                    str.replace (str.find(vardef), vardef.size (), varval);
                }
                else
                {
                    std::string message;
                    
                    if (line > 0)
                    {
                        message = _("environment variable ") + varname
                            + _(" not set in line ")
                            + Murx::Utils::int_to_string (line+1);
                    }                    
                    else
                    {
                        message = _("environment variable ")
                            + varname + _(" not set");
                    }
                    
                    throw (Murx::Exceptions::EnvironmentVariableNotSet (message));
                }
            }
            catch (std::out_of_range& e)
            {
                throw Murx::Exceptions::Exception
                    (_("expand_env_vars: out_of_range caught"));
            }
            catch (std::exception& e)
            {
                throw Murx::Exceptions::Exception
                    (_("expand_env_vars: stl exception caught"));
            }
        }
    }
}


void Murx::Utils::expand_path (std::string& str, unsigned int line)
{
    expand_cmd_subst (str,line);
    expand_env_vars  (str,line);
}


std::string Murx::Utils::int_to_string (int value)
{
    std::string tmp;
    std::ostringstream ostr;
    
    ostr << value;
    tmp = ostr.str ();
    
    return tmp;
}


void Murx::Utils::normalize_subject (std::string& str)
{
    // pos = 8: skip "Subject:"
    
    for (unsigned int pos = 8; pos < str.length (); pos++)
    {
        if (!isalnum(str[pos]) && !isblank(str[pos]))
        {
            str.erase(pos,1);
            --pos;   // we removed one char, check this position again
        }
    }
}


void Murx::Utils::remove_leading_blanks (std::string& str)
{
    unsigned int pos = 0;
    
    while (pos < str.length () && isspace (str[pos]))
    {
        pos++;
    }
    
    if (pos)
    {
        str.erase (0,pos);
    }
}


void Murx::Utils::remove_trailing_blanks (std::string& str)
{
    int len = static_cast<int> (str.length ()) - 1;
    int pos = len;
    
    while (pos >= 0 && isspace (str[pos]))
    {
        pos--;
    }
    
    if (pos < len)
    {
        str.erase (pos+1,len - pos);
    }
}


void Murx::Utils::split (const std::string& str,
                         const char sep,
                         std::vector<std::string>& vec)
{
    unsigned int pos = 0;   // start position
    unsigned int new_pos;
    
    while (pos < str.length ())
    {
        new_pos = str.find (sep, pos);
        
        if (new_pos > str.length ())
            new_pos = str.length ();
        
        // get next word out of string
        std::string substring = str.substr (pos, new_pos - pos);
        
        // remove leading and trailing blanks
        trim (substring);
        
        // add the word to the vector
        vec.push_back (substring);
        
        // set new search position
        pos = new_pos + 1;
    }
}


void Murx::Utils::tolower (std::string& str)
{
    transform (str.begin(), str.end(), str.begin(), ::tolower);
}


void Murx::Utils::trim (std::string& str)
{
    remove_leading_blanks  (str);
    remove_trailing_blanks (str);
}
