// murx.cxx - command line user interface of murx
//
// (c) 2004 by Kai Hildebrandt, Thomas Stauss
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------

#include <string>
#include <vector>
#include <iostream>
#include <exception>
//#include <stdexcept>
#include <csignal>   // SIGINT
#include "i18n.hxx"
#include "Logger.hxx"
#include "Match.hxx"
#include "MatchData.hxx"
#include "Preferences.hxx"
#include "Session.hxx"
#include "Exceptions.hxx"
#include "Header.hxx"
#include "Utils.hxx"

#include "SignalHandler.hxx"
#include "SIGINTHandler.hxx"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

extern "C"
{
#ifdef HAVE_GETOPT_H
#include <getopt.h>
#else
#include "getopt.h"
#endif
}

#include "murx.hxx"

#ifndef logger_logfile
#define logger_logfile Murx::Logger::get_instance(Murx::Logger::Logfile)
#endif

#ifndef logger_headerfile
#define logger_headerfile Murx::Logger::get_instance(Murx::Logger::Headerfile)
#endif


// additional definitions
struct option longopts[] =
{
    { "help", no_argument, 0, 'h' },
    { "headerfile", required_argument, 0, 'H' },
    { "logfile", required_argument, 0, 'L' },
    { "murxrc", required_argument, 0, 'M' },
    { "no-action", no_argument, 0, 'n' },
    { "return-value", no_argument, 0, 'r' },
    { "test", no_argument, 0, 't' },
    { "verbose", required_argument, 0, 'v' },
    { "version", no_argument, 0, 'V' },
    { 0, 0, 0, 0 }
};

// prototypes
static void show_headers (Murx::Header& header);
static void show_matches (Murx::Header& header);


int main (int argc, char** argv)
{
    int         option;   // value of command line options
    std::string rcfile;

    // set default values

    // set default location of rcfile
    rcfile = getenv("HOME");
    rcfile += "/.murxrc";

    std::string  logfile    = "";
    std::string  headerfile = "";
    bool         testmode   = false;
    unsigned int loglevel   = 0;
    bool         no_action  = false;

    // internally used flags
    bool         loglevel_set       = false;
    bool         return_remaining   = false;

    // error occurred while filtering?
    bool         was_error        = false;

    int          return_value     = Murx::SUCCESS;

    // this only has to be done once for initialization
#if ENABLE_NLS
#ifdef HAVE_SETLOCALE
    setlocale(LC_ALL, "");
#endif /* HAVE_SETLOCALE */
    bindtextdomain(PACKAGE, LOCALEDIR);
    textdomain(PACKAGE);
#endif /* ENABLE_NLS */

    // parse command line options
    while ((option = getopt_long(argc, argv, "hH:L:M:nrtv:V", longopts, (int*) 0)) != EOF)
    {
        switch (option)
        {
        case 'h':
            std::cout << _("Murx filters e-mail and removes spam in one ");
            std::cout << _("or many POP and IMAP accounts.") << std::endl;
            std::cout << std::endl;
            std::cout << _("Usage: ") << PACKAGE_NAME << _(" [OPTION]...")
                      << std::endl;
            std::cout << std::endl;
            std::cout << _("If a long option shows an argument as mandatory, ");
            std::cout << _("then ") << std::endl;
            std::cout << _("it is mandatory for the equivalent short option ");
            std::cout << _("also.") << std::endl;
            std::cout << std::endl;
            std::cout << _("Options:") << std::endl;
            std::cout << _("  -h, --help                 ");
            std::cout << _("Display this help information") << std::endl;
            std::cout << _("  -H, --headerfile=FILE      ");
            std::cout << _("Specify headerfile location") << std::endl;
            std::cout << _("  -L, --logfile=FILE         ");
            std::cout << _("Specify logfile location") << std::endl;
            std::cout << _("  -M, --murxrc=FILE          ");
            std::cout << _("Specify rcfile location") << std::endl;
            std::cout << _("  -n, --no-action            ");
            std::cout << _("Read rcfile and report errors") << std::endl;
            std::cout << _("  -r, --return-value         ");
            std::cout << _("Return amount of remaining messages on server") << std::endl;
            std::cout << _("  -t, --test                 ");
            std::cout << _("Simulate filter action (do not delete, move or add score)") << std::endl;
            std::cout << _("  -v, --verbose=LEVEL        ");
            std::cout << _("Specify level of verbosity") << std::endl;
            std::cout << _("  -V, --version              ");
            std::cout << _("Display version information") << std::endl;
            std::cout << std::endl;
            std::cout << _("Report bugs to ");
            std::cout << "<murx-devel@lists.sourceforge.net>." << std::endl;

            exit(Murx::SUCCESS);

        case 'H':
            headerfile = static_cast<std::string> (optarg);
            break;

        case 'L':
            logfile = static_cast<std::string> (optarg);
            break;

        case 'M':
            rcfile = static_cast<std::string> (optarg);
            break;

        case 'n':
            no_action = true;
            break;

        case 'r':
            return_remaining = true;
            break;

        case 't':
            testmode = true;
            break;

        case 'v':
            loglevel = static_cast<unsigned int> (atoi(optarg));
            loglevel_set = true;
            break;

        case 'V':
            std::cout << PACKAGE_NAME << " " << PACKAGE_VERSION << std::endl;
            std::cout << std::endl;
            std::cout
                << "Copyright (c) 2004-2007  Kai Hildebrandt, Thomas Stauss"
                << std::endl;
            std::cout << std::endl;
            std::cout << _("This is free software; see the source for copying")
                      << std::endl;
            std::cout << _("conditions.  There is NO warranty; not even for")
                      << std::endl;
            std::cout << _("MERCHANTABILITY or FITNESS FOR A PARTICULAR")
                      << std::endl;
            std::cout << _("PURPOSE.") << std::endl;

            exit(Murx::SUCCESS);

        default:
            // command line option not recognised

            std::cerr << _("Try '") << argv[0]
                 << _(" --help' for more information.") << std::endl;

            exit(Murx::IO_ERROR);
        }
    }

    // reset log message
    std::string logmessage = "";

    try
    {
        // Murx main program
        //
        // The variable rcfile contains either the default
        // location ${HOME}/.murxrc or the filename given by
        // the command line.
        //
        // The session object parses all options of rcfile.
        // Some of the options could be overwritten by the
        // command line so we correct these options before
        // we use them.
        // 
        Murx::Session session(rcfile);

        // parser run completed, print debug messages and exit
        // -> this is useful for rcfile debugging
        if (no_action)
        {
            std::cerr << PACKAGE_NAME << " " << PACKAGE_VERSION
                      << _(" finished successfully.") << std::endl;
            exit(Murx::SUCCESS);
        }

        // This is a singleton object of Murx::Preferences.
        // It is created in Murx::Session and will only be allocated once.
        // After the session object is destroyed we also can
        // not use the preferences object any more.
        //
        Murx::Preferences &preferences = *Murx::Preferences::get_instance();


        // No logfile given by the command line, get it from
        // the preferences object (could also be empty).
        // If a logfile given on the command line, store its
        // filename in preferences.
        if (logfile.empty())
            logfile = preferences.get_logfile();
        else
            preferences.set_logfile (logfile);

        // get the filename to which we should log headers to
        if (headerfile.empty())
            headerfile = preferences.get_headerfile();
        else
            preferences.set_headerfile (headerfile);


        // No loglevel given by the command line, get it from
        // the preferences object (default: 3). If loglevel
        // was set via command line store it in preferences.
        if (!loglevel_set)
        {
            loglevel = preferences.get_loglevel();
        }
        else
        {   // used by show_headers(...)
            preferences.set_loglevel(loglevel);
        }


        // Set the global testmode
        preferences.set_testmode(preferences.get_testmode() || testmode);


        // install a signal handler for SIGINT
        Murx::SIGINTHandler sigint_handler;

        Murx::SignalHandler::get_instance()->register_handler
            (SIGINT, &sigint_handler);


        // Some strings for creation of a logmessage
        std::string current_account_name = "";
        std::string action               = "";
        std::string reason               = "";
        

        // loglevel for a message
        unsigned int level = 0;


        // print program name and version
        logmessage = static_cast<std::string> (PACKAGE_NAME) + " "
            + static_cast<std::string> (PACKAGE_VERSION)
            + _(" started (rcfile is ") + rcfile + ")";
        logger_logfile->log(logmessage, Murx::Logger::NOTICE);


        // mailfilter loop
        while (1)
        {
            // Temporary object to store all information of
            // a checked mailheader (and probably body part)
            // including matched strings and mail size.
            Murx::Header header;

            try
            {
                if (session.filter_next_mail (header) == false)
                    break;
            }
            catch (Murx::Exceptions::NoFingerprint &e)
            {
                std::vector<std::string> logmessage;

                logmessage.push_back
                    (_("warning: no fingerprint defined for account \"")
                     + session.get_current_accountname()
                     + _("\" (skipped)"));
                logmessage.push_back
                    (_("         fingerprints of received server certificate are:"));
                logmessage.push_back
                    ("         SHA1: \"" + e.fingerprint_sha1 + "\"");
                logmessage.push_back
                    ("         MD5 : \"" + e.fingerprint_md5 + "\"");

                for (unsigned int i = 0; i < logmessage.size(); i++)
                    logger_logfile->log(logmessage[i], Murx::Logger::WARNING);

                // remember that there was an error
                was_error = true;

                if (return_remaining)
                    return_value = 0;
                else
                    return_value = Murx::AUTH_ERROR;
                
                continue;
            }
            catch (Murx::Exceptions::FingerprintChanged &e)
            {
                std::vector<std::string> logmessage;

                logmessage.push_back
                    (_("warning: fingerprint changed for account \"")
                     + session.get_current_accountname()
                     + _("\" (skipped)"));
                logmessage.push_back
                    (_("         fingerprints of received server certificate are:"));
                logmessage.push_back
                    ("         SHA1: \"" + e.fingerprint_sha1 + "\"");
                logmessage.push_back
                    ("         MD5 : \"" + e.fingerprint_md5 + "\"");

                for (unsigned int i = 0; i < logmessage.size(); i++)
                    logger_logfile->log(logmessage[i], Murx::Logger::WARNING);

                // remember that there was an error
                was_error = true;

                if (return_remaining)
                    return_value = 0;
                else
                    return_value = Murx::AUTH_ERROR;

                continue;
            }
            catch (Murx::Exceptions::LoginError &e)
            {
                std::string logmessage = 
                    _("warning: login failure in account \"")
                    + session.get_current_accountname() + "\"";
                logger_logfile->log(logmessage, Murx::Logger::WARNING);

                // remember that there was an error
                was_error = true;

                if (return_remaining)
                    return_value = 0;
                else
                    return_value = Murx::AUTH_ERROR;

                continue;
            }
            catch (Murx::Exceptions::Protocol &e)
            {
                std::vector<std::string> logmessage;

                logmessage.push_back
                    (_("warning: some error occured while communication with server of account \"")
                     + session.get_current_accountname() + "\"");
                logmessage.push_back
                    (_("         error was: ") + e.what());

                for (unsigned int i = 0; i < logmessage.size(); i++)
                    logger_logfile->log(logmessage[i], Murx::Logger::WARNING);

                // remember that there was an error
                was_error = true;

                if (return_remaining)
                    return_value = 0;
                else
                    return_value = Murx::NETWORK_ERROR;

                continue;
            }
            catch (Murx::Exceptions::Network &e)
            {
                std::vector<std::string> logmessage;

                logmessage.push_back
                    (_("warning: an error occured while filtering of account \"")
                     + session.get_current_accountname()
                     + "\" (skipped)");
                logmessage.push_back
                    (_("         error was: ") + e.what());

                for (unsigned int i = 0; i < logmessage.size(); i++)
                    logger_logfile->log(logmessage[i], Murx::Logger::WARNING);

                // remember that there was an error
                was_error = true;

                if (return_remaining)
                    return_value = 0;
                else
                    return_value = Murx::NETWORK_ERROR;

                continue;
            }
            


            if (session.get_current_accountname() != current_account_name
                && session.get_current_accountname() != "")
            {
                // Print out account name and count of messages
                // in the mailbox.

                current_account_name = session.get_current_accountname();

                logmessage = _("querying account \"")
                    + current_account_name + "\" ("
                    + Murx::Utils::int_to_string(session.get_current_message_count())
                    + _(" messages)");
                logger_logfile->log(logmessage, Murx::Logger::INFO);

                // some additional info (statistics)
                logmessage = _("checking ")
                    + Murx::Utils::int_to_string(session.get_current_account().get_allows().size())
                    + _(" ALLOW, ")
                    + Murx::Utils::int_to_string(session.get_current_account().get_denies().size())
                    + _(" DENY, ")
                    + Murx::Utils::int_to_string(session.get_current_account().get_movetos().size())
                    + _(" MOVETO and ")
                    + Murx::Utils::int_to_string(session.get_current_account().get_scores().size())
                    + _(" SCORE filters on this account.");
                logger_logfile->log(logmessage, Murx::Logger::INFO);
            }
            


            if (header.get_action() == session.ACCOUNT_FILTERING_FINISHED)
            {
                // remember count of remaining mails as return value
                return_value = session.get_remaining_message_count();

                logmessage = _("there are ")
                    + Murx::Utils::int_to_string(return_value)
                    + _(" messages left on account \"")
                    + session.get_current_accountname() +"\"";
                logger_logfile->log(logmessage, Murx::Logger::INFO);

                // We have to logout from the actual mailserver to delete
                // all spam mails to avoid a network timeout while filtering
                // the other accounts
                try
                {
                    session.finish_current_account();
                }
                catch (Murx::Exceptions::Exception &e)
                {
                    // TODO: Thomas
                    std::cout << "Murx::Exception: " << e.what() << std::endl;

                    was_error = true;
                }
                
                continue;
            }
            else if (header.get_action() == session.ALLOW)
            {
                // delete the message if global testmode is not set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    action = _("approved ");
                    level  = Murx::Logger::INFO;
                }
                else
                {
                    action = _("not approved (test) ");
                    level  = Murx::Logger::WARNING;
                }

                reason = _("ALLOW rule");
            }
            else if (header.get_action() == session.DELETE_MAXSIZE_ALLOW)
            {
                // delete the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("approved and deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("approved but not deleted (test) ");
                }

                // set reason and level for the log message
                reason = _("mail size of ")
                    + Murx::Utils::int_to_string(header.get_size())
                    + _(" exceeded MAXSIZE_ALLOW of ")
                    + Murx::Utils::int_to_string(preferences.get_maxsize_allow());
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.DELETE_DENYRULE)
            {
                // delete the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not deleted (test) ");
                }

                reason = _("DENY rule");
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.DELETE_MAXSIZE_DENY)
            {
                // delete the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not deleted (test) ");
                }

                reason = _("mail size of ")
                    + Murx::Utils::int_to_string(header.get_size())
                    + _(" exceeded MAXSIZE_DENY of ")
                    + Murx::Utils::int_to_string(preferences.get_maxsize_deny());
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.DELETE_DUPLICATE)
            {
                // delete the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not deleted (test) ");
                }

                reason = _("message was duplicate, ID: ")
                    + header.get_message_id();
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.DELETE_NONCONFORMANT)
            {
                // delete the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not deleted (test) ");
                }

                reason = _("message was not rfc2822-conformant");
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.DELETE_MAXLENGTH_EXCEEDED)
            {
                // delete the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not deleted (test) ");
                }

                reason = _("maxlength exceeded for at least one header line");
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.MOVETO)
            {
                // move the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // move mail
                    try
                    {
                        session.move_mail(header.get_account(),
                                          header.get_id(),
                                          header.get_folder());

                        action = _("moved to folder \"")
                            + header.get_folder() + "\" ";
                        reason = _("MOVETO rule");
                    }
                    catch (Murx::Exceptions::FolderDoesNotExist &e)
                    {
                        action = _("not moved (error) ");
                        reason = _("folder \"") + header.get_folder()
                            + _("\" does not exist");
                    }
                    catch (Murx::Exceptions::CommandNotAvailable &e)
                    {
                        std::string logmessage = _("warning: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::WARNING);

                        continue;
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not moved (test) ");
                    reason = _("MOVETO rule");
                }

                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.MOVETO_NONCONFORMANT)
            {
                // move the message if it is not marked for testing
                // or the global testmode is set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // move mail
                    try
                    {
                        session.move_mail(header.get_account(),
                                          header.get_id(),
                                          header.get_folder());

                        action = _("moved to folder \"")
                            + header.get_folder() + "\" ";
                        reason = _("MOVETO rule");
                    }
                    catch (Murx::Exceptions::FolderDoesNotExist &e)
                    {
                        action = _("not moved (error) ");
                        reason = _("folder \"") + header.get_folder()
                            + _("\" does not exist");
                    }
                    catch (Murx::Exceptions::CommandNotAvailable &e)
                    {
                        std::string logmessage = _("warning: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::WARNING);

                        continue;
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not moved (test) ");
                    reason = _("MOVETO rule (mail was not rfc2822-conformant)");
                }

                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.SCORE_DELETE)
            {
                // delete the message if it global testmode is not set
                
                if (!preferences.get_testmode() && !header.get_testmode())
                {
                    // delete mail
                    try
                    {
                        session.delete_mail(header.get_account(),
                                            header.get_id());

                        action = _("deleted ");
                    }
                    catch (Murx::Exceptions::Protocol &e)
                    {
                        std::string logmessage = _("error: ") + e.what();
                        logger_logfile->log(logmessage, Murx::Logger::ERROR);

                        continue;
                    }
                }
                else
                {
                    action = _("not deleted (test) ");
                }

                reason = _("SCORE ")
                    + Murx::Utils::int_to_string(header.get_total_score())
                    + _(" exceeded HIGHSCORE of ")
                    + Murx::Utils::int_to_string(preferences.get_highscore());
                level  = Murx::Logger::WARNING;
            }
            else if (header.get_action() == session.NORULE)
            {
                // at this point no ALLOW, DENY or MOVETO rule matched and the
                // HIGHSCORE was not reached for this message

                // show matches for the message
                //
                // it is also possible, that a message had some matches but they
                // are not shown (SCORE 0). This is because there are negative
                // rules but no complete filter matched and the special bit for
                // SHOW_ALL_MATCHES is not set.
                
                if (header.get_match_list().size() > 0)
                {
                    logmessage = _("matches for message: ")
                        + header.get_from() + ": "
                        + header.get_subject() + ", "
                        + header.get_date() + ". [" + "SCORE "
                        + Murx::Utils::int_to_string(header.get_total_score())
                        + "]";
                    logger_logfile->log(logmessage, Murx::Logger::SHOW_MATCHES);

                    // log all matches
                    show_matches (header);
                }

                // log the header
                show_headers (header);

                continue;
            }


            // log message: <action> <from>: <subject>, <date>. [<why>]

            // check for empty fields and 
            logmessage = action
                + ((! header.get_from().empty())? header.get_from() : "[ empty ]")
                + ": "
                + ((! header.get_subject().empty())? header.get_subject() : "[ empty ]")
                + ", "
                + ((! header.get_date().empty())? header.get_date() : "[ empty ]")
                + ". [" + reason + "]";
            logger_logfile->log(logmessage, level);

            // log all matches
            show_matches (header);

            // log the header
            show_headers (header);

            // --> filter_next_mail()

        }  // while (1)

        
        // we have to logout from all mailservers to delete
        // all spam mails before deletion of the Session object
        try
        {
            session.finish();
        }
        catch (Murx::Exceptions::Exception &e)
        {
            // TODO: Thomas
            std::cerr << static_cast<std::string> (PACKAGE_NAME)
                      << ": Caught Murx::Exception: " << e.what() << std::endl;

            was_error = true;
        }


        if (was_error)
        {
            logmessage = static_cast<std::string> (PACKAGE_NAME) + " "
                + static_cast<std::string> (PACKAGE_VERSION)
                + _(" finished with errors");

            if (return_remaining)
            {
                // if a return value of remaining mails is desired don't
                // return a negative error value because the shell does not
                // support it. We simply return 0 that a probable script will
                // not get confused.
                //
                return_value = 0;
            }
        }
        else
        {
            logmessage = static_cast<std::string> (PACKAGE_NAME) + " "
                + static_cast<std::string> (PACKAGE_VERSION)
                + _(" finished successfully");

            // the return value is set to the remaining messages on the
            // server. If this isn't desired, we change the return value to
            // SUCCESS (0).
            if (!return_remaining)
                return_value = Murx::SUCCESS;
        }

        // write a final log message
        logger_logfile->log(logmessage, Murx::Logger::NOTICE);
    }
    catch (Murx::Exceptions::IOError &e)
    {
        // included rcfile, logfile or headerfile not found

        // at this point it is not possible to log to a file so
        // we just print out the error message.
        std::cerr << static_cast<std::string> (PACKAGE_NAME) << ": " <<
            _("error: ") << e.what() << std::endl;

        if (return_remaining)
            return_value = 0;
        else
            return_value = Murx::IO_ERROR;
    }
    catch (Murx::Exceptions::Parser &e)
    {
        // catch all exceptions thrown by the rcfile parser.
        // they are all critical and lead to program termination.

        // at this point it is not possible to log to a file so
        // we just print out the error message.
        std::cerr << static_cast<std::string> (PACKAGE_NAME) << ": " <<
            _("error: ") << e.what() << std::endl;

        if (return_remaining)
            return_value = 0;
        else
            return_value = Murx::PARSE_ERROR;
    }
    catch (Murx::Exceptions::Exception &e)
    {
        std::cerr << static_cast<std::string> (PACKAGE_NAME)
                  << ": Uncaught Murx::Exception: \"" << e.what() << "\""
                  << std::endl;
    }
    catch (std::exception &e)
    {
        std::cerr << static_cast<std::string> (PACKAGE_NAME)
                  << ": Uncaught STL exception: \"" << e.what() << "\""
                  << std::endl;
    }
    catch (...)
    {
        std::cerr << static_cast<std::string> (PACKAGE_NAME)
                  << ": Uncaught an unknown exception!"
                  << std::endl << std::endl;
    }
    
    // close logfiles
    Murx::Logger::destroy_all();

    // a return value not equal zero is either an error or the 
    // amount of remaining mails on the server if the -r command line
    // switch is used.
    //
    // keep in mind that a return value of 0 with the -r option can be
    // either "no remaining mails" or an error.
    //
    return return_value;
}


static void show_headers (Murx::Header& header)
{
    if (Murx::Preferences::get_instance()->get_loglevel()
        & Murx::Logger::SHOW_HEADERS)
    {
        std::string logmessage;
        const std::vector<std::string>& hl = header.get_header_lines();
        
        // log header lines
        for (unsigned int i = 0; i < hl.size(); i++)
        {
            logmessage = hl.at(i);

            // it is not necessary to use the log(msg,lvl)-method because we
            // already checked the special bit and we do not want a date
            // in front of any header line
            logger_headerfile->log_header(logmessage);
        }

        // log body lines if available
        const std::vector<std::string>& bl = header.get_body_lines();
        
        if (bl.size() > 0)
        {
            // header/body separator
            logmessage = "";
            logger_headerfile->log_header(logmessage);

            for (unsigned int i = 0; i < bl.size(); i++)
            {
                logmessage = bl.at(i);
                
                // it is not necessary to use the log(msg,lvl)-method because we
                // already checked the special bit and we do not want a date
                // in front of any header line
                logger_headerfile->log_header(logmessage);
            }
            
        }

        // insert an blank line as separator between two headers
        logger_headerfile->log_header("");
    }
}


static void show_matches (Murx::Header& header)
{
    std::string logmessage;

    const std::vector<Murx::Match>& match_list = header.get_match_list();

    // the last line with match for comparation with
    // actual line with match so that we can output
    // this line only once
    int last_line = -1;

    // >>> outer loop
    // vector match_list -> match
    for (unsigned int i = 0; i < match_list.size(); i++)
    {
        const Murx::Match& current_match = match_list[i];

        // create the log message
        logmessage = _("filter");
        if (!current_match.was_complete_match())
        {
            logmessage += _(" not");
        }
        logmessage += _(" matched");
        logmessage += " [" + current_match.get_action() + "]:";

        // Logger decides if we want to see partly matched filters (special bit)
        if (current_match.was_complete_match())
            logger_logfile->log(logmessage, Murx::Logger::SHOW_MATCHES);
        else
            logger_logfile->log(logmessage, Murx::Logger::SHOW_ALL_MATCHES);

        // don't show partly matched filters:
        // - either the filter matched completely
        // - or some patterns matched and SHOW_ALL_MATCHES is set
        if (current_match.was_complete_match()
            || (Murx::Preferences::get_instance()->get_loglevel()
                & Murx::Logger::SHOW_ALL_MATCHES)
            ==    Murx::Logger::SHOW_ALL_MATCHES)
        {
            // >>> inner loop
            // match -> match_data
            for (unsigned int j = 0; j < current_match.size(); j++)
            {
                const Murx::MatchData& current_data = current_match.at(j);
                
                std::string expr = current_data.get_expression();
                
                if (current_data.get_negativity())
                {
                    logmessage = _("negative pattern \"")
                        + expr + _("\" does not match");
                    logger_logfile->log(logmessage, Murx::Logger::SHOW_MATCHES);
                }
                else
                {
                    std::string line;
                    std::string where;
                    
                    // Does the pattern match in header or body of the mail?
                    if (current_data.get_is_header_line())
                    {
                        // header
                        line = header.get_header_lines().at(current_data.get_line());
                        where = _("header line");
                    }
                    else
                    {
                        // body
                        line = header.get_body_lines().at(current_data.get_line());
                        where = _("body line");
                    }
                    
                    unsigned int start_pos  = current_data.get_start_pos();
                    unsigned int count      = current_data.get_count();
                    std::string  normalized = "";
                    
                    
                    // log the matched header or body line (un-normalized)
                    if (last_line != static_cast<int> (current_data.get_line()))
                    {
                        // remember this line index
                        last_line = current_data.get_line();
                        
                        // print out the line
                        logmessage = _("pattern(s) matched in ")
                            + where + " \"" + line + "\"";
                        logger_logfile->log(logmessage, Murx::Logger::SHOW_LINE_WITH_MATCH);
                    }
                    
                    
                    if (current_data.get_matched_normalized_subject())
                    {
                        Murx::Utils::normalize_subject(line);
                        normalized = _(" (normalized)");
                    }
                    
                    // log pattern and matched substring
                    try
                    {
                        logmessage = _("pattern \"")
                            + expr + _("\" matched substring \"")
                            + line.substr(start_pos,count) + "\""
                            + normalized; // + _(" in ") + where;
                        logger_logfile->log(logmessage, Murx::Logger::SHOW_MATCHES);
                    }
                    catch (std::exception &e)
                    {
                        // TODO
                        std::cerr << "main(): caught STL exception at substr():"
                                  << " line.length() = "
                                  << Murx::Utils::int_to_string (line.length())
                                  << "; start_pos = "
                                  << Murx::Utils::int_to_string (start_pos)
                                  << "; count = "
                                  << Murx::Utils::int_to_string (count)
                                  << std::endl;
                    }
                }
            } // <<< inner loop
        }
    } // <<< outer loop

    
    // show local size matches
    //
    // NOTE:
    //
    // There are two different variable names for size limiation:
    // minsize and maxsize. But for logging it is always named maxsize.
    //
    if (header.get_maxsize() != 0)
    {
        // TODO: hmmm... probably "does not reach local minsize"?
        logmessage = _("mail size ")
            + Murx::Utils::int_to_string(header.get_size())
            +  _(" does not exceed local maxsize of ")
            + Murx::Utils::int_to_string(header.get_maxsize());
        logger_logfile->log(logmessage, Murx::Logger::SHOW_MATCHES);
    }

    if (header.get_minsize() != 0)
    {
        logmessage = _("mail size ")
            + Murx::Utils::int_to_string(header.get_size())
            +  _(" exceeds local maxsize of ")
            + Murx::Utils::int_to_string(header.get_minsize());
        logger_logfile->log(logmessage, Murx::Logger::SHOW_MATCHES);
    }
}
