// testRCParser.cxx - part of the murx project
//                    standalone parser for rcfile debugging
//
// (c) 2004-2006 by Kai Hildebrandt
//
// ---------------------------------------------------------------------
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
// USA.
//
// ---------------------------------------------------------------------


#include <vector>
#include <fstream>
#include <iostream>
#include "i18n.hxx"
#include "Filter.hxx"
#include "Score.hxx"
#include "MoveTo.hxx"
#include "Account.hxx"
#include "Preferences.hxx"
#include "Exceptions.hxx"
#include "RCParser.hxx"
#include "Utils.hxx"


void print         ();
void print_filters (const Murx::Account& account);


int main (int argc, char* argv[])
{
	Murx::Preferences*	preferences;
	Murx::RCParser*		rcparser;

    if (argc != 2)
    {
        std::cerr << _("usage: ") << argv[0] << _(" <rcfile>") << std::endl;
        return -1;
    }

    // loop how many times?
    int loop = 1;

    while (loop)
    {
        // open the rcfile
        std::filebuf i_rcfile;
        
        std::cout << _("reading rcfile ") << argv[1] << std::endl;
        i_rcfile.open(argv[1], std::ios::in);
        
        if (!i_rcfile.is_open())
        {
            std::cout << argv[0] << _(": could not open rcfile \"") << argv[1]
                      << "\"" << std::endl;
            return -2;
        }
        
        std::istream i_stream(&i_rcfile);

        preferences = Murx::Preferences::get_instance();
        rcparser    = new Murx::RCParser(&i_stream);
        
        //std::cout << "testRCParser started" << std::endl;
        std::cout << _("parsing...") << std::endl;
        
        try
        {
            rcparser->parse(preferences);
        }
        catch (Murx::Exceptions::SyntaxError& e)
        {
            std::cerr << _("syntax error: ") << e.what() << std::endl;
            throw;
        }
        catch (Murx::Exceptions::Exception& e)
        {
            std::cerr << _("exception: ") << e.what() << std::endl;
            throw;
        }

        // no longer needed
        delete rcparser;

        std::cout << _("closing rcfile") << std::endl;
        i_rcfile.close();

        std::cout << _("rcfile seems to be syntactically correct :-)")
                  << std::endl;

        print();

        Murx::Preferences::destroy();

        // next
        loop--;

        std::cout << "----------------------------------------------------------------------" << std::endl;
    }
}

void print ()
{
	Murx::Preferences& preferences = *Murx::Preferences::get_instance();
    const std::vector<Murx::Account>& accounts = preferences.get_accounts();
        
	std::cout << std::endl << std::endl;
	std::cout << _("contents of rcfile:") << std::endl;

	std::cout << std::endl;
	std::cout << _("defined accounts:") << std::endl;

    if (accounts.size() > 0)
    {
        for (unsigned int i = 0; i < accounts.size(); i++)
        {
            accounts[i].print();
            std::cout << std::endl;
        }
    }
    else
    {
        std::cout << _("warning: no accounts defined! Probably this is an INCLUDE-file?")
                  << std::endl;
    }
    std::cout << std::endl;

	std::cout << _("defined global settings:") << std::endl;
	std::cout << _("(hint: some of this options can be overridden via command line options.") << std::endl;
	std::cout << _(" check the manpage murx(1) or call murx --help to see how this works.)") << std::endl;
	std::cout << std::endl;
	std::cout << _("*** also keep this in mind if you are using a wrapper script ***") << std::endl;
	std::cout << std::endl;

	preferences.print();

	std::cout << std::endl;

    for (unsigned int current_account = 0; current_account < accounts.size(); current_account++)
    {
        std::cout << _("listing all filters defined for account ")
                  << accounts[current_account].get_name()
                  << ":" << std::endl << std::endl;

        print_filters (accounts[current_account]);
    }

	std::cout << std::endl;
}



void print_filters(const Murx::Account& account)
{
    const std::vector<Murx::Filter>& allows  = account.get_allows();
    const std::vector<Murx::Filter>& denies  = account.get_denies();
    const std::vector<Murx::MoveTo>& movetos = account.get_movetos();
    const std::vector<Murx::Score>&  scores  = account.get_scores();


    if (allows.size() > 0)
        std::cout << _("ALLOW filters:") << std::endl;

    for (unsigned int j = 0; j < allows.size(); j++)
    {
        allows[j].print();
    }


    if (denies.size() > 0)
        std::cout << _("DENY filters:") << std::endl;
    
    for (unsigned int j = 0; j < denies.size(); j++)
    {
        denies[j].print();
    }


    if (movetos.size() > 0)
        std::cout << _("MOVETO filters:") << std::endl;
    
    for (unsigned int j = 0; j < movetos.size(); j++)
    {
        movetos[j].print();
    }


    if (scores.size() > 0)
        std::cout << _("SCORE filters:") << std::endl;
     
    for (unsigned int j = 0; j < scores.size(); j++)
    {
        scores[j].print();
    }
}
